# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from pathlib import Path
from typing import Literal

import matplotlib.pyplot as plt
import numpy as np
import torch
from torch_geometric.data import HeteroData
from torch_geometric.data.storage import EdgeStorage
from torch_geometric.data.storage import NodeStorage

from anemoi.graphs.plotting.prepare import compute_node_adjacencies
from anemoi.graphs.plotting.prepare import get_edge_attribute_dims
from anemoi.graphs.plotting.prepare import get_node_attribute_dims

LOGGER = logging.getLogger(__name__)


def plot_distribution_node_attributes(graph: HeteroData, out_file: str | Path | None = None) -> None:
    """Figure with the distribution of the node attributes.

    Each row represents a node type and each column an attribute dimension.
    """
    num_nodes = len(graph.node_types)
    attr_dims = get_node_attribute_dims(graph)
    plot_distribution_attributes(graph.node_items(), num_nodes, attr_dims, "Node", out_file)


def plot_distribution_edge_attributes(graph: HeteroData, out_file: str | Path | None = None) -> None:
    """Figure with the distribution of the edge attributes.

    Each row represents a edge type and each column an attribute dimension.
    """
    num_edges = len(graph.edge_types)
    attr_dims = get_edge_attribute_dims(graph)
    plot_distribution_attributes(graph.edge_items(), num_edges, attr_dims, "Edge", out_file)


def plot_distribution_node_derived_attributes(graph, outfile: str | Path | None = None):
    """Figure with the distribution of the node derived attributes.

    Each row represents a node type and each column an attribute dimension.
    """
    node_adjs = {}
    node_attr_dims = {}
    for source_name, _, target_name in graph.edge_types:
        node_adj_tensor = compute_node_adjacencies(graph, source_name, target_name)
        node_adj_tensor = torch.from_numpy(node_adj_tensor.reshape((node_adj_tensor.shape[0], -1)))
        node_adj_key = f"# edges from {source_name}"

        # Store node adjacencies
        if target_name in node_adjs:
            node_adjs[target_name] = node_adjs[target_name] | {node_adj_key: node_adj_tensor}
        else:
            node_adjs[target_name] = {node_adj_key: node_adj_tensor}

        # Store attribute dimension
        if node_adj_key not in node_attr_dims:
            node_attr_dims[node_adj_key] = node_adj_tensor.shape[1]

    node_adj_list = [(k, v) for k, v in node_adjs.items()]

    plot_distribution_attributes(node_adj_list, len(node_adjs), node_attr_dims, "Node", outfile)


def plot_distribution_attributes(
    graph_items: NodeStorage | EdgeStorage,
    num_items: int,
    attr_dims: dict,
    item_type: Literal["Edge", "Node"],
    out_file: str | Path | None = None,
) -> None:
    """Figure with the distribution of the node and edge attributes.

    Each row represents a node or edge type and each column an attribute dimension.
    """
    dim_attrs = sum(attr_dims.values())

    if dim_attrs == 0:
        LOGGER.warning("No edge attributes found in the graph.")
        return None

    # Define the layout
    _, axs = plt.subplots(num_items, dim_attrs, figsize=(10 * num_items, 10))
    if num_items == dim_attrs == 1:
        axs = np.array([[axs]])
    elif axs.ndim == 1:
        axs = axs.reshape(num_items, dim_attrs)

    for i, (item_name, item_store) in enumerate(graph_items):
        for j, (attr_name, attr_values) in enumerate(attr_dims.items()):
            for dim in range(attr_values):
                if attr_name in item_store:
                    axs[i, j + dim].hist(item_store[attr_name][:, dim].float(), bins=50)

                    axs[i, j + dim].set_ylabel("".join(item_name).replace("to", " --> "))
                    axs[i, j + dim].set_title(attr_name if attr_values == 1 else f"{attr_name}_{dim}")
                    if i == num_items - 1:
                        axs[i, j + dim].set_xlabel(attr_name if attr_values == 1 else f"{attr_name}_{dim}")
                else:
                    axs[i, j + dim].set_axis_off()

    plt.suptitle(f"{item_type} Attributes distribution", fontsize=14)
    plt.savefig(out_file)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest
import torch
from omegaconf import OmegaConf
from torch_geometric.data import HeteroData

from anemoi.graphs.nodes.attributes import SphericalAreaWeights
from anemoi.graphs.nodes.attributes import UniformWeights
from anemoi.graphs.nodes.builders import from_file


def test_init(mocker, mock_anemoi_dataset_cutout):
    """Test AnemoiDatasetNodes initialization with cutout."""
    mocker.patch("anemoi.datasets.open_dataset", return_value=mock_anemoi_dataset_cutout)
    node_builder = from_file.AnemoiDatasetNodes(
        OmegaConf.create({"cutout": ["lam.zarr", "global.zarr"]}), name="test_nodes"
    )

    assert isinstance(node_builder, from_file.BaseNodeBuilder)
    assert isinstance(node_builder, from_file.AnemoiDatasetNodes)


def test_register_nodes(mocker, mock_anemoi_dataset_cutout):
    """Test AnemoiDatasetNodes register correctly the nodes with cutout operation."""
    mocker.patch("anemoi.datasets.open_dataset", return_value=mock_anemoi_dataset_cutout)

    node_builder = from_file.AnemoiDatasetNodes(
        OmegaConf.create({"cutout": ["lam.zarr", "global.zarr"]}), name="test_nodes"
    )
    graph = HeteroData()

    graph = node_builder.register_nodes(graph)

    assert graph["test_nodes"].x is not None
    assert isinstance(graph["test_nodes"].x, torch.Tensor)
    assert graph["test_nodes"].x.shape == (mock_anemoi_dataset_cutout.num_nodes, 2)
    assert graph["test_nodes"].node_type == "AnemoiDatasetNodes"


@pytest.mark.parametrize("attr_class", [UniformWeights, SphericalAreaWeights])
def test_register_attributes(mocker, mock_anemoi_dataset_cutout, graph_with_nodes: HeteroData, attr_class):
    """Test AnemoiDatasetNodes register correctly the weights with cutout operation."""
    mocker.patch("anemoi.datasets.open_dataset", return_value=mock_anemoi_dataset_cutout)

    node_builder = from_file.AnemoiDatasetNodes(
        OmegaConf.create({"cutout": ["lam.zarr", "global.zarr"]}), name="test_nodes"
    )
    config = {"test_attr": {"_target_": f"anemoi.graphs.nodes.attributes.{attr_class.__name__}"}}

    graph = node_builder.register_attributes(graph_with_nodes, config)

    assert graph["test_nodes"]["test_attr"] is not None
    assert isinstance(graph["test_nodes"]["test_attr"], torch.Tensor)
    assert graph["test_nodes"]["test_attr"].shape[0] == graph["test_nodes"].x.shape[0]



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from collections.abc import Callable

import numpy as np
import torch
from scipy.sparse import coo_matrix
from torch_geometric.data.storage import NodeStorage

from anemoi.graphs.generate.transforms import latlon_rad_to_cartesian

LOGGER = logging.getLogger(__name__)


class NodeMaskingMixin:
    """Mixin class for masking source/target nodes when building edges."""

    def get_node_coordinates(
        self, source_nodes: NodeStorage, target_nodes: NodeStorage
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Get the node coordinates."""
        source_coords, target_coords = source_nodes.x, target_nodes.x

        if self.source_mask_attr_name is not None:
            source_coords = source_coords[source_nodes[self.source_mask_attr_name].squeeze()]

        if self.target_mask_attr_name is not None:
            target_coords = target_coords[target_nodes[self.target_mask_attr_name].squeeze()]

        return source_coords.to(self.device), target_coords.to(self.device)

    def get_cartesian_node_coordinates(
        self, source_nodes: NodeStorage, target_nodes: NodeStorage
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Get cartesian coordinates of the source and target nodes.

        This method handles also the masking of source/target nodes if specified.

        Attributes
        ----------
        source_nodes : NodeStorage
            The source node storage
        target_nodes : NodeStorage
            The target node storage

        Return
        ------
        source_coords : torch.Tensor
            Source coordinates.
        target_coords : torch.Tensor
            Target coordinates.
        """
        source_coords, target_coords = self.get_node_coordinates(source_nodes, target_nodes)
        source_coords = latlon_rad_to_cartesian(source_coords)
        target_coords = latlon_rad_to_cartesian(target_coords)
        return source_coords, target_coords

    @staticmethod
    def get_unmasking_mapping(mask: torch.Tensor) -> Callable:
        masked_indices = np.where(mask.squeeze().cpu())[0]
        mapper = dict(zip(range(len(masked_indices)), masked_indices))
        return np.vectorize(mapper.get)

    def undo_masking_adj_matrix(self, adj_matrix, source_nodes: NodeStorage, target_nodes: NodeStorage):
        """Undo masking for adjacency matrix, remapping indices to original node indices.

        Arguments
        ---------
        adj_matrix
            Adjacency matrix
        source_nodes : NodeStorage
            Source node storage.
        target_nodes : NodeStorage
            Target node storage.

        Returns
        -------
        np.ndarray
            Remapped adj_matrix with original node indices.
        """
        if self.target_mask_attr_name is not None:
            target_node_mapping = NodeMaskingMixin.get_unmasking_mapping(mask=target_nodes[self.target_mask_attr_name])
            adj_matrix.row = target_node_mapping(adj_matrix.row)

        if self.source_mask_attr_name is not None:
            source_node_mapping = NodeMaskingMixin.get_unmasking_mapping(mask=source_nodes[self.source_mask_attr_name])
            adj_matrix.col = source_node_mapping(adj_matrix.col)

        if self.source_mask_attr_name is not None or self.target_mask_attr_name is not None:
            true_shape = target_nodes.x.shape[0], source_nodes.x.shape[0]
            adj_matrix = coo_matrix((adj_matrix.data, (adj_matrix.row, adj_matrix.col)), shape=true_shape)

        return adj_matrix

    def undo_masking_edge_index(
        self, edge_index: torch.Tensor, source_nodes: NodeStorage, target_nodes: NodeStorage
    ) -> torch.Tensor:
        """Undo masking for edge_index matrix, remapping indices to original node indices.

        Arguments
        ---------
        edge_index : torch.Tensor
            2 x N tensor of edge indices (target, source).
        source_nodes : NodeStorage
            Source node storage.
        target_nodes : NodeStorage
            Target node storage.

        Returns
        -------
        torch.Tensor
            Remapped edge_index with original node indices.
        """
        # Remap source indices (row 0)
        if self.source_mask_attr_name is not None:
            source_node_mapping = NodeMaskingMixin.get_unmasking_mapping(mask=source_nodes[self.source_mask_attr_name])
            edge_index[0] = torch.from_numpy(source_node_mapping(edge_index[0].cpu().numpy())).to(edge_index.device)

        # Remap target indices (row 1)
        if self.target_mask_attr_name is not None:
            target_node_mapping = NodeMaskingMixin.get_unmasking_mapping(mask=target_nodes[self.target_mask_attr_name])
            edge_index[1] = torch.from_numpy(target_node_mapping(edge_index[1].cpu().numpy())).to(edge_index.device)

        return edge_index



# (C) Copyright 2025- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest
from torch_geometric.data import HeteroData

from anemoi.graphs.edges import HEALPixMultiScaleEdges
from anemoi.graphs.nodes import HEALPixNodes


class TestMultiScaleEdgesInit:
    def test_init(self):
        """Test MultiScaleEdges initialisation."""
        assert isinstance(HEALPixMultiScaleEdges("test_nodes", "test_nodes", None), HEALPixMultiScaleEdges)
        assert isinstance(
            HEALPixMultiScaleEdges("test_nodes", "test_nodes", scale_resolutions=4), HEALPixMultiScaleEdges
        )
        assert isinstance(
            HEALPixMultiScaleEdges("test_nodes", "test_nodes", scale_resolutions=[1, 2, 3]), HEALPixMultiScaleEdges
        )

    @pytest.mark.parametrize("scale_resolutions", [0, -1, [0], [-1], "invalid"])
    def test_fail_init_invalid_scale_resolutions(self, scale_resolutions):
        """Test MultiScaleEdges initialisation with invalid scale_resolutions."""
        with pytest.raises(AssertionError):
            HEALPixMultiScaleEdges("test_nodes", "test_nodes", scale_resolutions=scale_resolutions)

    def test_fail_init_diff_nodes(self):
        """Test MultiScaleEdges initialisation with invalid nodes."""
        with pytest.raises(AssertionError):
            HEALPixMultiScaleEdges("test_nodes", "test_nodes2", None)


class TestHEALPixMultiScaleEdgesTransform:

    @pytest.fixture()
    def healpix_graph(self) -> HeteroData:
        """Return a HeteroData object with HEALPixMultiScaleEdges."""
        graph = HeteroData()
        graph = HEALPixNodes(1, "test_tri_nodes").update_graph(graph, {})
        graph["fail_nodes"].x = [1, 2, 3]
        graph["fail_nodes"].node_type = "FailNodes"
        return graph

    def test_transform_same_src_dst_tri_nodes(self, healpix_graph: HeteroData):
        """Test HEALPixMultiScaleEdges update method."""

        edges = HEALPixMultiScaleEdges("test_tri_nodes", "test_tri_nodes", None)
        graph = edges.update_graph(healpix_graph)
        assert ("test_tri_nodes", "to", "test_tri_nodes") in graph.edge_types

    @pytest.mark.parametrize("scale_resolutions", [1, [1], [1, 2], None])
    def test_transform_with_scale_resolutions(self, healpix_graph: HeteroData, scale_resolutions):
        """Test HEALPixMultiScaleEdges with different scale_resolutions configurations."""
        edges = HEALPixMultiScaleEdges("test_tri_nodes", "test_tri_nodes", scale_resolutions=scale_resolutions)
        graph = edges.update_graph(healpix_graph)

        assert ("test_tri_nodes", "to", "test_tri_nodes") in graph.edge_types
        assert len(graph[("test_tri_nodes", "to", "test_tri_nodes")].edge_index) > 0
        assert graph[("test_tri_nodes", "to", "test_tri_nodes")].edge_index.dim() == 2

    def test_transform_fail_nodes(self, healpix_graph: HeteroData):
        """Test MultiScaleEdges update method with wrong node type."""
        edges = HEALPixMultiScaleEdges("fail_nodes", "fail_nodes", None)
        with pytest.raises(AssertionError):
            edges.update_graph(healpix_graph)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from .builders.from_file import AnemoiDatasetNodes
from .builders.from_file import LimitedAreaNPZFileNodes
from .builders.from_file import NPZFileNodes
from .builders.from_file import TextNodes
from .builders.from_file import XArrayNodes
from .builders.from_healpix import HEALPixNodes
from .builders.from_healpix import LimitedAreaHEALPixNodes
from .builders.from_icon import ICONCellGridNodes
from .builders.from_icon import ICONMultiMeshNodes
from .builders.from_reduced_gaussian import ReducedGaussianGridNodes
from .builders.from_refined_icosahedron import HexNodes
from .builders.from_refined_icosahedron import LimitedAreaHexNodes
from .builders.from_refined_icosahedron import LimitedAreaTriNodes
from .builders.from_refined_icosahedron import StretchedTriNodes
from .builders.from_refined_icosahedron import TriNodes
from .builders.from_vectors import LatLonNodes

__all__ = [
    "AnemoiDatasetNodes",
    "NPZFileNodes",
    "TriNodes",
    "HexNodes",
    "HEALPixNodes",
    "LatLonNodes",
    "LimitedAreaHEALPixNodes",
    "LimitedAreaNPZFileNodes",
    "LimitedAreaTriNodes",
    "LimitedAreaHexNodes",
    "ReducedGaussianGridNodes",
    "StretchedTriNodes",
    "ICONMultiMeshNodes",
    "ICONCellGridNodes",
    "TextNodes",
    "XArrayNodes",
]



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from importlib.util import find_spec

import numpy as np
import torch
from scipy.sparse import coo_matrix
from sklearn.neighbors import NearestNeighbors
from torch_geometric.data import HeteroData
from torch_geometric.data.storage import NodeStorage
from torch_geometric.nn import radius

from anemoi.graphs import EARTH_RADIUS
from anemoi.graphs.edges.builders.base import BaseDistanceEdgeBuilders
from anemoi.graphs.utils import get_grid_reference_distance

TORCH_CLUSTER_AVAILABLE = find_spec("torch_cluster") is not None


LOGGER = logging.getLogger(__name__)


class CutOffEdges(BaseDistanceEdgeBuilders):
    """Computes cut-off based edges and adds them to the graph.

    It uses as reference the target nodes.

    Attributes
    ----------
    source_name : str
        The name of the source nodes.
    target_name : str
        The name of the target nodes.
    cutoff_factor : float | None
        Factor to multiply the grid reference distance to get the cut-off radius.
        Mutually exclusive with cutoff_distance_km.
    cutoff_distance_km : float | None
        Cutoff radius in kilometers. Mutually exclusive with cutoff_factor.
    source_mask_attr_name : str | None
        The name of the source mask attribute to filter edge connections.
    target_mask_attr_name : str | None
        The name of the target mask attribute to filter edge connections.
    max_num_neighbours : int
        The maximum number of nearest neighbours to consider when building edges.

    Methods
    -------
    register_edges(graph)
        Register the edges in the graph.
    register_attributes(graph, config)
        Register attributes in the edges of the graph.
    update_graph(graph, attrs_config)
        Update the graph with the edges.
    """

    def __init__(
        self,
        source_name: str,
        target_name: str,
        cutoff_factor: float | None = None,
        cutoff_distance_km: float | None = None,
        source_mask_attr_name: str | None = None,
        target_mask_attr_name: str | None = None,
        max_num_neighbours: int = 64,
    ) -> None:
        super().__init__(source_name, target_name, source_mask_attr_name, target_mask_attr_name)

        # Validate that exactly one of cutoff_factor or cutoff_distance_km is provided
        if cutoff_factor is None and cutoff_distance_km is None:
            raise ValueError("Either cutoff_factor or cutoff_distance_km must be provided.")
        if cutoff_factor is not None and cutoff_distance_km is not None:
            raise ValueError("cutoff_factor and cutoff_distance_km are mutually exclusive. Provide only one.")

        if cutoff_factor is not None:
            assert isinstance(cutoff_factor, (int, float)), "Cutoff factor must be a float."
            assert cutoff_factor > 0, "Cutoff factor must be positive."

        if cutoff_distance_km is not None:
            assert isinstance(cutoff_distance_km, (int, float)), "Cutoff distance must be a float."
            assert cutoff_distance_km > 0, "Cutoff distance must be positive."

        assert isinstance(max_num_neighbours, int), "Number of nearest neighbours must be an integer."
        assert max_num_neighbours > 0, "Number of nearest neighbours must be positive."

        self.cutoff_factor = cutoff_factor
        self.cutoff_distance_km = cutoff_distance_km
        self.max_num_neighbours = max_num_neighbours

    @staticmethod
    def get_reference_distance(nodes: NodeStorage, mask_attr_name: torch.Tensor | None = None) -> float:
        """Compute the reference distance.

        Parameters
        ----------
        nodes : NodeStorage
            The nodes.
        mask_attr_name : str
            The mask attribute name.

        Returns
        -------
        float
            The nodes reference distance.
        """
        if mask_attr_name is not None:
            # If masking nodes, we have to recompute the grid reference distance only over the masked nodes
            mask = nodes[mask_attr_name]
            _grid_reference_distance = get_grid_reference_distance(nodes.x, mask)
        else:
            _grid_reference_distance = nodes["_grid_reference_distance"]

        return _grid_reference_distance

    def get_cutoff_radius(self, graph: HeteroData):
        """Compute the cut-off radius.

        The cut-off radius is computed either as:
        - The product of the target nodes reference distance and the cut-off factor, or
        - Directly from the cutoff distance in kilometers.

        Parameters
        ----------
        graph : HeteroData
            The graph.

        Returns
        -------
        float
            The cut-off radius in Cartesian coordinates on unit sphere.
        """
        if self.cutoff_distance_km is not None:
            # Convert km to Cartesian distance on unit sphere
            # For small distances: Cartesian distance ≈ great circle distance (radians)
            # radians = km / EARTH_RADIUS
            radius = self.cutoff_distance_km / EARTH_RADIUS
        else:
            # Use factor-based approach
            reference_dist = CutOffEdges.get_reference_distance(
                graph[self.target_name], mask_attr_name=self.target_mask_attr_name
            )
            radius = reference_dist * self.cutoff_factor
        return radius

    def prepare_node_data(self, graph: HeteroData) -> tuple[NodeStorage, NodeStorage]:
        """Prepare node information and get source and target nodes."""
        self.radius = self.get_cutoff_radius(graph)
        return super().prepare_node_data(graph)

    def _compute_edge_index_pyg(self, source_coords: torch.Tensor, target_coords: torch.Tensor) -> torch.Tensor:
        edge_index = radius(source_coords, target_coords, r=self.radius, max_num_neighbors=self.max_num_neighbours)
        edge_index = torch.flip(edge_index, [0])

        return edge_index

    def _crop_to_max_num_neighbours(self, adjmat):
        """Remove neighbors exceeding the maximum allowed limit."""
        nodes_to_drop = np.maximum(np.bincount(adjmat.row) - self.max_num_neighbours, 0)
        if (num_nodes_to_drop := nodes_to_drop.sum()) == 0:
            return adjmat

        LOGGER.info(
            "%s is removing %d because they exceed the maximum allowed number of neighbors (%d) for each target node.",
            self.__class__.__name__,
            num_nodes_to_drop,
            self.max_num_neighbours,
        )

        # Vectorized approach: sort edges by (row, distance) to group by node
        # no repeated O(nnz) scans in a loop
        sort_idx = np.lexsort((adjmat.data, adjmat.row))
        sorted_rows = adjmat.row[sort_idx]

        # Find where each row starts and ends
        row_changes = np.concatenate(([0], np.where(np.diff(sorted_rows) != 0)[0] + 1, [len(sorted_rows)]))

        # Compute rank of each edge within its row
        edge_rank_in_row = np.zeros(len(sorted_rows), dtype=int)
        for i in range(len(row_changes) - 1):
            start, end = row_changes[i], row_changes[i + 1]
            edge_rank_in_row[start:end] = np.arange(end - start)

        # Keep edges where rank < max_num_neighbours (smallest distances are first due to sorting)
        mask_sorted = edge_rank_in_row < self.max_num_neighbours

        # Map back to original order
        mask = np.zeros(adjmat.nnz, dtype=bool)
        mask[sort_idx] = mask_sorted

        # Define the new sparse matrix
        return coo_matrix((adjmat.data[mask], (adjmat.row[mask], adjmat.col[mask])), shape=adjmat.shape)

    def _compute_adj_matrix_sklearn(self, source_coords: torch.Tensor, target_coords: torch.Tensor) -> torch.Tensor:
        nearest_neighbour = NearestNeighbors(metric="euclidean", n_jobs=4)
        nearest_neighbour.fit(source_coords.cpu())
        adj_matrix = nearest_neighbour.radius_neighbors_graph(
            target_coords.cpu(), radius=self.radius, mode="distance"
        ).tocoo()

        adj_matrix = self._crop_to_max_num_neighbours(adj_matrix)
        return adj_matrix

    def compute_edge_index(self, source_nodes: NodeStorage, target_nodes: NodeStorage) -> torch.Tensor:
        """Get the adjacency matrix for the cut-off method.

        Parameters
        ----------
        source_nodes : NodeStorage
            The source nodes.
        target_nodes : NodeStorage
            The target nodes.

        Returns
        -------
        torch.Tensor of shape (2, num_edges)
            The adjacency matrix.
        """
        radius_km = self.radius * EARTH_RADIUS
        if self.cutoff_distance_km is not None:
            LOGGER.info(
                "Using CutOff-Edges (with radius = %.1f km [direct]) between %s and %s.",
                radius_km,
                self.source_name,
                self.target_name,
            )
        else:
            LOGGER.info(
                "Using CutOff-Edges (with radius = %.1f km [factor=%.2f]) between %s and %s.",
                radius_km,
                self.cutoff_factor,
                self.source_name,
                self.target_name,
            )
        return super().compute_edge_index(source_nodes=source_nodes, target_nodes=target_nodes)


class ReversedCutOffEdges(CutOffEdges):
    """Computes cut-off based edges and adds them to the graph.

    It uses as reference the source nodes.

    Attributes
    ----------
    source_name : str
        The name of the source nodes.
    target_name : str
        The name of the target nodes.
    cutoff_factor : float
        Factor to multiply the grid reference distance to get the cut-off radius.
    source_mask_attr_name : str | None
        The name of the source mask attribute to filter edge connections.
    target_mask_attr_name : str | None
        The name of the target mask attribute to filter edge connections.
    max_num_neighbours : int
        The maximum number of nearest neighbours to consider when building edges.

    Methods
    -------
    register_edges(graph)
        Register the edges in the graph.
    register_attributes(graph, config)
        Register attributes in the edges of the graph.
    update_graph(graph, attrs_config)
        Update the graph with the edges.
    """

    def get_cartesian_node_coordinates(
        self, source_nodes: NodeStorage, target_nodes: NodeStorage
    ) -> tuple[torch.Tensor, torch.Tensor]:
        source_coords, target_coords = super().get_cartesian_node_coordinates(source_nodes, target_nodes)
        return target_coords, source_coords

    def get_cutoff_radius(self, graph: HeteroData):
        """Compute the cut-off radius.

        The cut-off radius is computed either as:
        - The product of the source nodes reference distance and the cut-off factor, or
        - Directly from the cutoff distance in kilometers.

        Parameters
        ----------
        graph : HeteroData
            The graph.

        Returns
        -------
        float
            The cut-off radius in Cartesian coordinates on unit sphere.
        """
        if self.cutoff_distance_km is not None:
            # Convert km to Cartesian distance on unit sphere
            # For small distances: Cartesian distance ≈ great circle distance (radians)
            # radians = km / EARTH_RADIUS
            radius = self.cutoff_distance_km / EARTH_RADIUS
        else:
            # Use factor-based approach
            reference_dist = CutOffEdges.get_reference_distance(
                graph[self.source_name], mask_attr_name=self.source_mask_attr_name
            )
            radius = reference_dist * self.cutoff_factor
        return radius

    def undo_masking_adj_matrix(self, adj_matrix, source_nodes: NodeStorage, target_nodes: NodeStorage):
        adj_matrix = adj_matrix.T
        return super().undo_masking_adj_matrix(adj_matrix, source_nodes, target_nodes)

    def undo_masking_edge_index(
        self, edge_index: torch.Tensor, source_nodes: NodeStorage, target_nodes: NodeStorage
    ) -> torch.Tensor:
        edge_index = torch.flip(edge_index, [0])
        return super().undo_masking_edge_index(edge_index, source_nodes, target_nodes)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from pathlib import Path

import numpy as np
import torch
from omegaconf import DictConfig
from omegaconf import OmegaConf
from torch_geometric.data import HeteroData

from anemoi.graphs.generate.masks import KNNAreaMaskBuilder
from anemoi.graphs.nodes.builders.base import BaseNodeBuilder

LOGGER = logging.getLogger(__name__)


class AnemoiDatasetNodes(BaseNodeBuilder):
    """Nodes from an anemoi dataset.

    Attributes
    ----------
    dataset : str | DictConfig
        The dataset.

    Methods
    -------
    get_coordinates()
        Get the lat-lon coordinates of the nodes.
    register_nodes(graph, name)
        Register the nodes in the graph.
    register_attributes(graph, name, config)
        Register the attributes in the nodes of the graph specified.
    update_graph(graph, name, attrs_config)
        Update the graph with new nodes and attributes.
    """

    def __init__(self, dataset: DictConfig, name: str) -> None:
        LOGGER.info("Reading the dataset from %s.", dataset)
        self.dataset = dataset if isinstance(dataset, str) else OmegaConf.to_container(dataset)
        super().__init__(name)
        self.hidden_attributes = BaseNodeBuilder.hidden_attributes | {"dataset"}

    def get_coordinates(self) -> torch.Tensor:
        """Get the coordinates of the nodes.

        Returns
        -------
        torch.Tensor of shape (num_nodes, 2)
            A 2D tensor with the coordinates, in radians.
        """
        from anemoi.datasets import open_dataset

        dataset = open_dataset(self.dataset)
        return self.reshape_coords(dataset.latitudes, dataset.longitudes)


class TextNodes(BaseNodeBuilder):
    """Nodes from text file.

    Attributes
    ----------
    dataset : str | Path
        The path including filename to txt file containing the coordinates of the nodes.
    idx_lon : int
        The index of the longitude in the dataset.
    idx_lat : int
        The index of the latitude in the dataset.
    """

    def __init__(self, dataset: str | Path, name: str, idx_lon: int = 0, idx_lat: int = 1) -> None:
        LOGGER.info("Reading the dataset from %s.", dataset)
        self.dataset = dataset
        self.idx_lon = idx_lon
        self.idx_lat = idx_lat
        super().__init__(name)

    def get_coordinates(self) -> torch.Tensor:
        """Get the coordinates of the nodes.

        Returns
        -------
        torch.Tensor of shape (num_nodes, 2)
            A 2D tensor with the coordinates, in radians.
        """
        dataset = np.loadtxt(self.dataset)
        return self.reshape_coords(dataset[self.idx_lat, :], dataset[self.idx_lon, :])


class NPZFileNodes(BaseNodeBuilder):
    """Nodes from NPZ defined grids.

    Attributes
    ----------
    npz_file : str
        Path to the file.
    lat_key : str
        Name of the key of the latitude arrays.
    lon_key : str
        Name of the key of the latitude arrays.

    Methods
    -------
    get_coordinates()
        Get the lat-lon coordinates of the nodes.
    register_nodes(graph, name)
        Register the nodes in the graph.
    register_attributes(graph, name, config)
        Register the attributes in the nodes of the graph specified.
    update_graph(graph, name, attrs_config)
        Update the graph with new nodes and attributes.
    """

    def __init__(
        self,
        npz_file: str,
        name: str,
        lat_key: str = "latitudes",
        lon_key: str = "longitudes",
    ) -> None:
        """Initialize the NPZFileNodes builder.

        The builder suppose the grids are stored in files with the name `grid-{resolution}.npz`.

        Parameters
        ----------
        npz_file : str
            The path to the file.
        name : str
            Name of the nodes to be added.
        lat_key : str, optional
            Name of the key of the latitude arrays. Defaults to "latitudes".
        lon_key : str, optional
            Name of the key of the latitude arrays. Defaults to "longitudes".
        """
        self.npz_file = Path(npz_file)
        self.lat_key = lat_key
        self.lon_key = lon_key
        super().__init__(name)

    def get_coordinates(self) -> torch.Tensor:
        """Get the coordinates of the nodes.

        Returns
        -------
        torch.Tensor of shape (num_nodes, 2)
            A 2D tensor with the coordinates, in radians.
        """
        assert self.npz_file.exists(), f"{self.__class__.__name__}.file does not exists: {self.npz_file}"
        grid_data = np.load(self.npz_file)
        coords = self.reshape_coords(grid_data[self.lat_key], grid_data[self.lon_key])
        return coords


class LimitedAreaNPZFileNodes(NPZFileNodes):
    """Nodes from NPZ defined grids using an area of interest."""

    def __init__(
        self,
        npz_file: str,
        reference_node_name: str,
        name: str,
        lat_key: str = "latitudes",
        lon_key: str = "longitudes",
        mask_attr_name: str | None = None,
        margin_radius_km: float = 100.0,
    ) -> None:
        super().__init__(npz_file, name, lat_key, lon_key)

        self.area_mask_builder = KNNAreaMaskBuilder(reference_node_name, margin_radius_km, mask_attr_name)

    def register_nodes(self, graph: HeteroData) -> None:
        self.area_mask_builder.fit(graph)
        return super().register_nodes(graph)

    def get_coordinates(self) -> np.ndarray:
        coords = super().get_coordinates()

        LOGGER.info(
            "Limiting the processor mesh to a radius of %.2f km from the output mesh.",
            self.area_mask_builder.margin_radius_km,
        )
        area_mask = self.area_mask_builder.get_mask(coords)

        LOGGER.info(
            "Dropping %d nodes from the processor mesh.",
            len(area_mask) - area_mask.sum(),
        )
        coords = coords[area_mask]

        return coords


class XArrayNodes(BaseNodeBuilder):
    """Class for creating graph nodes based on a xarray-compatible file format.

    Parameters
    ----------
    dataset : str
        Path to xarray compatible file (e.g., NetCDF or zarr) containing latitude and longitude variables.
    name : str
        Identifier to use for the nodes within the graph.
    lat_key : str, optional
        Variable name for latitude in the dataset (default: "lat").
    lon_key : str, optional
        Variable name for longitude in the dataset (default: "lon").

    Methods
    -------
    get_coordinates()
        Get the lat-lon coordinates of the nodes.
    register_nodes(graph, name)
        Register the nodes in the graph.
    register_attributes(graph, name, config)
        Register the attributes in the nodes of the graph specified.
    update_graph(graph, name, attrs_config)
        Update the graph with new nodes and attributes.
    """

    def __init__(self, dataset: str, name: str, lat_key: str = "lat", lon_key: str = "lon") -> None:

        super().__init__(name)
        self.dataset = dataset
        self.lat_key = lat_key
        self.lon_key = lon_key
        self.hidden_attributes = BaseNodeBuilder.hidden_attributes | {"dataset"}

    def get_coordinates(self) -> torch.Tensor:
        import xarray as xr

        ds = xr.open_dataset(self.dataset)

        for var in [self.lat_key, self.lon_key]:
            assert var in ds, f"Variable '{var}' not found in dataset."

        lat = ds[self.lat_key].values.flatten()
        lon = ds[self.lon_key].values.flatten()
        return self.reshape_coords(lat, lon)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest
import torch

from anemoi.graphs.normalise import NormaliserMixin


@pytest.mark.parametrize("norm", ["l1", "l2", "unit-max", "unit-range", "unit-std", "log1p"])
def test_normaliser(norm: str):
    """Test NormaliserMixin normalise method."""

    class Normaliser(NormaliserMixin):
        def __init__(self, norm):
            self.norm = norm
            self.norm_by_group = False

        def __call__(self, data):
            return self.normalise(data)

    normaliser = Normaliser(norm=norm)
    data = torch.rand(10, 5)
    normalised_data = normaliser(data)
    assert isinstance(normalised_data, torch.Tensor)
    assert normalised_data.shape == data.shape


@pytest.mark.parametrize("norm", ["l1", "l2", "unit-max", "unit-range", "unit-std", "log1p"])
def test_grouped_normaliser(norm: str):
    """Test NormaliserMixin normalise method."""

    class Normaliser(NormaliserMixin):
        def __init__(self, norm):
            self.norm = norm
            self.norm_by_group = True

        def __call__(self, data, index, num_groups):
            return self.normalise(data, index, num_groups)

    normaliser = Normaliser(norm=norm)
    data = torch.rand(10, 5)
    index = torch.tensor([0, 1, 2, 3, 4, 0, 1, 2, 3, 4])
    num_groups = 5
    normalised_data = normaliser(data, index, num_groups)
    assert isinstance(normalised_data, torch.Tensor)
    assert normalised_data.shape == data.shape


@pytest.mark.parametrize("norm", ["l3", "invalid"])
def test_normaliser_wrong_norm(norm: str):
    """Test NormaliserMixin normalise method."""

    class Normaliser(NormaliserMixin):
        def __init__(self, norm: str):
            self.norm = norm
            self.norm_by_group = False

        def __call__(self, data):
            return self.normalise(data)

    with pytest.raises(AssertionError):
        normaliser = Normaliser(norm=norm)
        data = torch.rand(10, 5)
        normaliser(data)


def test_normaliser_wrong_inheritance():
    """Test NormaliserMixin normalise method."""

    class Normaliser(NormaliserMixin):
        def __init__(self, attr):
            self.attr = attr

        def __call__(self, data):
            return self.normalise(data)

    with pytest.raises(AttributeError):
        normaliser = Normaliser(attr="attr_name")
        data = torch.rand(10, 5)
        normaliser(data)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import networkx as nx
import numpy as np
import pytest
from torch_geometric.data import HeteroData

from anemoi.graphs.edges import MultiScaleEdges
from anemoi.graphs.generate import tri_icosahedron
from anemoi.graphs.generate.masks import KNNAreaMaskBuilder
from anemoi.graphs.nodes import HexNodes
from anemoi.graphs.nodes import StretchedTriNodes
from anemoi.graphs.nodes import TriNodes


class TestMultiScaleEdgesInit:
    def test_init(self):
        """Test MultiScaleEdges initialization."""
        assert isinstance(MultiScaleEdges("test_nodes", "test_nodes", 1, None), MultiScaleEdges)
        assert isinstance(MultiScaleEdges("test_nodes", "test_nodes", 1, scale_resolutions=4), MultiScaleEdges)
        assert isinstance(MultiScaleEdges("test_nodes", "test_nodes", 1, scale_resolutions=[1, 2, 3]), MultiScaleEdges)

    @pytest.mark.parametrize("x_hops", [-0.5, "hello", None, -4])
    def test_fail_init_invalid_x_hops(self, x_hops: str):
        """Test MultiScaleEdges initialization with invalid x_hops."""
        with pytest.raises(AssertionError):
            MultiScaleEdges("test_nodes", "test_nodes", x_hops, None)

    @pytest.mark.parametrize("scale_resolutions", [-1, [-1, 0], "invalid"])
    def test_fail_init_invalid_scale_resolutions(self, scale_resolutions):
        """Test MultiScaleEdges initialization with invalid scale_resolutions."""
        with pytest.raises(AssertionError):
            MultiScaleEdges("test_nodes", "test_nodes", 1, scale_resolutions=scale_resolutions)

    def test_fail_init_diff_nodes(self):
        """Test MultiScaleEdges initialization with invalid nodes."""
        with pytest.raises(AssertionError):
            MultiScaleEdges("test_nodes", "test_nodes2", 0, None)


class TestMultiScaleEdgesTransform:

    @pytest.fixture()
    def tri_ico_graph(self) -> HeteroData:
        """Return a HeteroData object with MultiScaleEdges."""
        graph = HeteroData()
        graph = TriNodes(2, "test_tri_nodes").update_graph(graph, {})
        graph["fail_nodes"].x = [1, 2, 3]
        graph["fail_nodes"].node_type = "FailNodes"
        return graph

    @pytest.fixture()
    def hex_ico_graph(self) -> HeteroData:
        """Return a HeteroData object with TriNodes."""
        graph = HeteroData()
        graph = HexNodes(1, "test_hex_nodes").update_graph(graph, {})
        graph["fail_nodes"].x = [1, 2, 3]
        graph["fail_nodes"].node_type = "FailNodes"
        return graph

    def test_transform_same_src_dst_tri_nodes(self, tri_ico_graph: HeteroData):
        """Test MultiScaleEdges update method."""

        edges = MultiScaleEdges("test_tri_nodes", "test_tri_nodes", 1, None)
        graph = edges.update_graph(tri_ico_graph)
        assert ("test_tri_nodes", "to", "test_tri_nodes") in graph.edge_types

    @pytest.mark.parametrize("edge_resolutions", [[0], [0, 1, 2], [0, 2], [2]])
    def test_fast_1_hop_method(selg, tri_ico_graph: HeteroData, edge_resolutions):
        nodes = tri_ico_graph["test_tri_nodes"]
        fast_edges = tri_icosahedron.add_1_hop_edges(
            nodes_coords_rad=nodes["x"],
            node_resolutions=nodes["_resolutions"],
            edge_resolutions=edge_resolutions,
            node_ordering=nodes["_node_ordering"],
            area_mask_builder=nodes.get("_area_mask_builder", None),
        )
        nx_graph = tri_icosahedron.add_edges_to_nx_graph(
            nodes["_nx_graph"],
            resolutions=edge_resolutions,
            x_hops=1,
            area_mask_builder=nodes.get("_area_mask_builder", None),
        )
        adjmat = nx.to_scipy_sparse_array(nx_graph, format="coo")
        slow_edges = np.stack([adjmat.col, adjmat.row], axis=0)
        # Order of the edges might be different, so lets make sets of the pairs
        fast_edge_pairs = set(map(tuple, fast_edges.T))
        slow_edge_pairs = set(map(tuple, slow_edges.T))
        assert fast_edge_pairs == slow_edge_pairs

    def test_transform_same_src_dst_hex_nodes(self, hex_ico_graph: HeteroData):
        """Test MultiScaleEdges update method."""

        edges = MultiScaleEdges("test_hex_nodes", "test_hex_nodes", 1, None)
        graph = edges.update_graph(hex_ico_graph)
        assert ("test_hex_nodes", "to", "test_hex_nodes") in graph.edge_types

    @pytest.mark.parametrize("scale_resolutions", [1, [1], [1, 2], None])
    def test_transform_with_scale_resolutions(self, tri_ico_graph: HeteroData, scale_resolutions):
        """Test MultiScaleEdges with different scale_resolutions configurations."""
        edges = MultiScaleEdges("test_tri_nodes", "test_tri_nodes", 1, scale_resolutions=scale_resolutions)
        graph = edges.update_graph(tri_ico_graph)

        assert ("test_tri_nodes", "to", "test_tri_nodes") in graph.edge_types
        assert len(graph[("test_tri_nodes", "to", "test_tri_nodes")].edge_index) > 0
        assert graph[("test_tri_nodes", "to", "test_tri_nodes")].edge_index.dim() == 2

    def test_transform_fail_nodes(self, tri_ico_graph: HeteroData):
        """Test MultiScaleEdges update method with wrong node type."""
        edges = MultiScaleEdges("fail_nodes", "fail_nodes", 1, None)
        with pytest.raises(ValueError):
            edges.update_graph(tri_ico_graph)


class TestMultiScaleEdgesStretched:

    @pytest.fixture()
    def tri_graph(self, mocker) -> HeteroData:
        """Return a HeteroData object with stretched Tri nodes."""
        graph = HeteroData()
        node_builder = StretchedTriNodes(4, 6, "hidden", None, None)
        node_builder.area_mask_builder = KNNAreaMaskBuilder("hidden", 400)
        node_builder.area_mask_builder.fit_coords(np.array([[0, 0]]))
        # We are considering a 400km radius circle centered at (0, 0) as the area of
        # interest for the stretched graph.

        mocker.patch.object(node_builder.area_mask_builder, "fit", return_value=None)

        graph = node_builder.update_graph(graph, {})
        return graph

    def test_edges(self, tri_graph: HeteroData):
        """Test MultiScaleEdges update method."""
        edges = MultiScaleEdges("hidden", "hidden", x_hops=1, scale_resolutions=None)
        graph = edges.update_graph(tri_graph)
        assert ("hidden", "to", "hidden") in graph.edge_types
        assert len(graph[("hidden", "to", "hidden")].edge_index) > 0

    @pytest.mark.parametrize("edge_resolutions", [[1], [0, 1, 2, 3, 4, 5, 6], [4, 6], [6]])
    def test_fast_1_hop_method(selg, tri_graph: HeteroData, edge_resolutions):
        nodes = tri_graph["hidden"]
        all_points_mask_builder = KNNAreaMaskBuilder("all_nodes", 1.0)
        all_points_mask_builder.fit_coords(nodes.x.detach().cpu().numpy())

        fast_edges = tri_icosahedron.add_1_hop_edges(
            nodes_coords_rad=nodes["x"],
            node_resolutions=nodes["_resolutions"],
            edge_resolutions=edge_resolutions,
            node_ordering=nodes["_node_ordering"],
            area_mask_builder=all_points_mask_builder,
        )

        nx_graph = tri_icosahedron.add_edges_to_nx_graph(
            nodes["_nx_graph"],
            resolutions=edge_resolutions,
            x_hops=1,
            area_mask_builder=all_points_mask_builder,
        )
        adjmat = nx.to_scipy_sparse_array(nx_graph, format="coo")
        slow_edges = np.stack([adjmat.col, adjmat.row], axis=0)

        # Order of the edges might be different, so lets make sets of the pairs
        fast_edge_pairs = set(map(tuple, fast_edges.T))
        slow_edge_pairs = set(map(tuple, slow_edges.T))
        assert fast_edge_pairs == slow_edge_pairs



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest
from torch_geometric.data import HeteroData

from anemoi.graphs.edges import CutOffEdges
from anemoi.graphs.edges import ReversedCutOffEdges


@pytest.mark.parametrize("edge_builder", [CutOffEdges, ReversedCutOffEdges])
def test_init(edge_builder):
    """Test CutOffEdges initialization."""
    edge_builder("test_nodes1", "test_nodes2", 0.5)


@pytest.mark.parametrize("edge_builder", [CutOffEdges, ReversedCutOffEdges])
@pytest.mark.parametrize("cutoff_factor", [-0.5, "hello"])
def test_fail_init_invalid_cutoff_factor(edge_builder, cutoff_factor: str):
    """Test CutOffEdges initialization with invalid cutoff_factor."""
    with pytest.raises(AssertionError):
        edge_builder("test_nodes1", "test_nodes2", cutoff_factor=cutoff_factor)


@pytest.mark.parametrize("edge_builder", [CutOffEdges, ReversedCutOffEdges])
def test_fail_init_no_params(edge_builder):
    """Test CutOffEdges initialization with neither cutoff_factor nor cutoff_distance_km."""
    with pytest.raises(ValueError, match="Either cutoff_factor or cutoff_distance_km must be provided"):
        edge_builder("test_nodes1", "test_nodes2")


@pytest.mark.parametrize("edge_builder", [CutOffEdges, ReversedCutOffEdges])
def test_fail_init_both_params(edge_builder):
    """Test CutOffEdges initialization with both cutoff_factor and cutoff_distance_km."""
    with pytest.raises(ValueError, match="mutually exclusive"):
        edge_builder("test_nodes1", "test_nodes2", cutoff_factor=0.5, cutoff_distance_km=500.0)


@pytest.mark.parametrize("edge_builder", [CutOffEdges, ReversedCutOffEdges])
def test_cutoff(edge_builder, graph_with_nodes: HeteroData):
    """Test CutOffEdges with cutoff_factor."""
    builder = edge_builder("test_nodes", "test_nodes", cutoff_factor=0.5)
    graph = builder.update_graph(graph_with_nodes)
    assert ("test_nodes", "to", "test_nodes") in graph.edge_types


@pytest.mark.parametrize("edge_builder", [CutOffEdges, ReversedCutOffEdges])
def test_cutoff_with_distance_km(edge_builder, graph_with_nodes: HeteroData):
    """Test CutOffEdges with cutoff_distance_km."""
    builder = edge_builder("test_nodes", "test_nodes", cutoff_distance_km=500.0)
    graph = builder.update_graph(graph_with_nodes)
    assert ("test_nodes", "to", "test_nodes") in graph.edge_types


@pytest.mark.parametrize("edge_builder", [CutOffEdges, ReversedCutOffEdges])
@pytest.mark.parametrize("cutoff_distance_km", [-500.0, "hello"])
def test_fail_init_invalid_cutoff_distance_km(edge_builder, cutoff_distance_km):
    """Test CutOffEdges initialization with invalid cutoff_distance_km."""
    with pytest.raises(AssertionError):
        edge_builder("test_nodes1", "test_nodes2", cutoff_distance_km=cutoff_distance_km)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import importlib
import logging

import torch
from torch_geometric.data.storage import NodeStorage

from anemoi.graphs.edges.builders.base import BaseEdgeBuilder
from anemoi.graphs.generate.multi_scale_edges import BaseIcosahedronEdgeStrategy

LOGGER = logging.getLogger(__name__)


class MultiScaleEdges(BaseEdgeBuilder):
    """Base class for multi-scale edges in the nodes of a graph.

    Attributes
    ----------
    source_name : str
        The name of the source nodes.
    target_name : str
        The name of the target nodes.
    x_hops : int
        Number of hops (in the refined icosahedron) between two nodes to connect
        them with an edge.
    scale_resolutions : int, list[int], optional
        Defines the refinement levels at which edges are computed. If an integer is provided, edges are computed for all
        levels up to and including that level. For instance, `scale_resolutions=4` includes edges at levels 0 through 4,
        whereas `scale_resolutions=[4]` only includes edges at level 4.

    Methods
    -------
    register_edges(graph)
        Register the edges in the graph.
    register_attributes(graph, config)
        Register attributes in the edges of the graph.
    update_graph(graph, attrs_config)
        Update the graph with the edges.

    Note
    ----
    `MultiScaleEdges` only supports computing the edges within a set of nodes built by an `Type[IcosahedronNodes]`.
    """

    def __init__(
        self,
        source_name: str,
        target_name: str,
        x_hops: int,
        scale_resolutions: int | list[int] | None = None,
        **kwargs,
    ):
        super().__init__(source_name, target_name)
        assert source_name == target_name, f"{self.__class__.__name__} requires source and target nodes to be the same."
        assert isinstance(x_hops, int), "Number of x_hops must be an integer"
        assert x_hops > 0, "Number of x_hops must be positive"
        self.x_hops = x_hops
        if isinstance(scale_resolutions, int):
            assert scale_resolutions >= 0, "The scale_resolutions argument only supports non-negative integers."
            scale_resolutions = list(range(1, scale_resolutions + 1))
        assert not isinstance(scale_resolutions, str), "The scale_resolutions argument is not valid."
        assert (
            scale_resolutions is None or min(scale_resolutions) >= 0
        ), "The scale_resolutions argument only supports non-negative integers."
        self.scale_resolutions = scale_resolutions

    @staticmethod
    def get_edge_builder_class(node_type: str) -> type[BaseIcosahedronEdgeStrategy]:
        # All node builders inheriting from IcosahedronNodes have an attribute multi_scale_edge_cls
        module = importlib.import_module("anemoi.graphs.nodes.builders.from_refined_icosahedron")
        node_cls = getattr(module, node_type, None)

        if node_cls is None:
            raise ValueError(f"Invalid node_type, {node_type}, for building multi scale edges.")

        # Instantiate the BaseIcosahedronEdgeStrategy based on the node type
        module_name = ".".join(node_cls.multi_scale_edge_cls.split(".")[:-1])
        class_name = node_cls.multi_scale_edge_cls.split(".")[-1]

        edge_builder_cls = getattr(importlib.import_module(module_name), class_name)
        return edge_builder_cls

    def compute_edge_index(self, source_nodes: NodeStorage, _target_nodes: NodeStorage) -> torch.Tensor:
        edge_builder_cls = MultiScaleEdges.get_edge_builder_class(source_nodes.node_type)

        # Get the refinement levels (or scales) at which to compute the neighbourhoods.
        scale_resolutions = self.scale_resolutions or source_nodes["_resolutions"]
        if self.scale_resolutions is not None and max(self.scale_resolutions) > max(source_nodes["_resolutions"]):
            LOGGER.warning(
                f"Some scale resolutions may be ignored because they are greater than the resolution of the nodes ({max(source_nodes['_resolutions'])})."
            )

        # Add edges
        edge_index = edge_builder_cls().get_edges(
            source_nodes,
            self.x_hops,
            scale_resolutions=scale_resolutions,
        )

        return torch.from_numpy(edge_index).to(torch.int32)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
import os
from pathlib import Path

import torch

from anemoi.graphs.plotting.displots import plot_distribution_edge_attributes
from anemoi.graphs.plotting.displots import plot_distribution_node_attributes
from anemoi.graphs.plotting.displots import plot_distribution_node_derived_attributes
from anemoi.graphs.plotting.interactive_2d_html import plot_interactive_nodes_2d
from anemoi.graphs.plotting.interactive_2d_html import plot_interactive_subgraph_2d
from anemoi.graphs.plotting.interactive_2d_html import plot_isolated_nodes_2d
from anemoi.graphs.plotting.interactive_3d_html import plot_interactive_graph_3d
from anemoi.graphs.processors.post_process import SubsetNodesInArea

LOGGER = logging.getLogger(__name__)


class GraphInspector:
    """Inspect the graph.

    Attributes
    ----------
    path: str | Path
        Path to the graph file.
    output_path: Path
        Path to the output directory where the plots will be saved.
    show_attribute_distributions: bool, optional
        Whether to show the distribution of the node and edge attributes.
    show_nodes: bool, optional
        Whether to show the interactive plots of the nodes.
    """

    def __init__(
        self,
        path: str | Path,
        output_path: Path,
        area: tuple[float, float, float, float] = None,
        show_attribute_distributions: bool | None = True,
        show_nodes: bool | None = False,
        **kwargs,
    ):
        self.path = path
        self.graph = torch.load(self.path, weights_only=False, map_location="cpu")
        self.area = area
        self.output_path = output_path
        self.show_attribute_distributions = show_attribute_distributions
        self.show_nodes = show_nodes

        if self.area is not None:
            self.graph = SubsetNodesInArea(nodes_name=list(self.graph.node_types), area=area).update_graph(self.graph)

        if isinstance(self.output_path, str):
            self.output_path = Path(self.output_path)

        os.makedirs(self.output_path, exist_ok=True)

        assert self.output_path.is_dir(), f"Path {self.output_path} is not a directory."
        assert os.access(self.output_path, os.W_OK), f"Path {self.output_path} is not writable."

    def inspect(self):
        """Run all the inspector methods."""

        LOGGER.info("Saving interactive 3d plot of the graph...")
        plot_interactive_graph_3d(self.graph, out_file=self.output_path / "graph.html")

        LOGGER.info("Saving interactive plots of isolated nodes ...")
        plot_isolated_nodes_2d(self.graph, self.output_path / "isolated_nodes.html")

        LOGGER.info("Saving interactive plots of subgraphs ...")
        for edges_subgraph in self.graph.edge_types:
            ofile = self.output_path / f"{edges_subgraph[0]}_to_{edges_subgraph[2]}.html"
            plot_interactive_subgraph_2d(self.graph, edges_subgraph, out_file=ofile)

        if self.show_attribute_distributions:
            LOGGER.info("Saving distribution plots of node ande edge attributes ...")
            plot_distribution_node_derived_attributes(self.graph, self.output_path / "distribution_node_adjancency.png")
            plot_distribution_edge_attributes(self.graph, self.output_path / "distribution_edge_attributes.png")
            plot_distribution_node_attributes(self.graph, self.output_path / "distribution_node_attributes.png")

        if self.show_nodes:
            LOGGER.info("Saving interactive plots of nodes ...")
            for nodes_name in self.graph.node_types:
                plot_interactive_nodes_2d(
                    self.graph, nodes_name, out_file=self.output_path / f"{nodes_name}_nodes.html"
                )



# (C) Copyright 2024- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from abc import ABC
from abc import abstractmethod

import torch
from torch_geometric.data.storage import NodeStorage

from anemoi.graphs.nodes.attributes.base_attributes import BooleanBaseNodeAttribute

LOGGER = logging.getLogger(__name__)
MaskAttributeType = str | type["BooleanBaseNodeAttribute"]


class BooleanOperation(BooleanBaseNodeAttribute, ABC):
    """Base class for boolean operations."""

    def __init__(self, masks: MaskAttributeType | list[MaskAttributeType]) -> None:
        super().__init__()
        assert masks is not None, f"{self.__class__.__name__} requires a valid masks argument."
        self.masks = masks if isinstance(masks, list) else [masks]

    @staticmethod
    def get_mask_values(mask: MaskAttributeType, nodes: NodeStorage, **kwargs) -> torch.Tensor:
        if isinstance(mask, str):
            assert mask in nodes, f"Nodes have no attribute named {mask}."
            attributes = nodes[mask]
            assert (
                attributes.dtype == torch.bool
            ), f"The mask attribute '{mask}' must be a boolean but is {attributes.dtype}."
            return attributes

        return mask.get_raw_values(nodes, **kwargs)

    @abstractmethod
    def reduce_op(self, masks: list[torch.Tensor]) -> torch.Tensor: ...

    def get_raw_values(self, nodes: NodeStorage, **kwargs) -> torch.Tensor:
        mask_values = [BooleanOperation.get_mask_values(mask, nodes, **kwargs) for mask in self.masks]
        return self.reduce_op(torch.stack(mask_values))


class BooleanNot(BooleanOperation):
    """Boolean NOT mask."""

    def reduce_op(self, masks: list[torch.Tensor]) -> torch.Tensor:
        assert len(self.masks) == 1, f"The {self.__class__.__name__} can only be aplied to one mask."
        return ~masks[0]


class BooleanAndMask(BooleanOperation):
    """Boolean AND mask."""

    def reduce_op(self, masks: list[torch.Tensor]) -> torch.Tensor:
        return torch.all(masks, dim=0)


class BooleanOrMask(BooleanOperation):
    """Boolean OR mask."""

    def reduce_op(self, masks: list[torch.Tensor]) -> torch.Tensor:
        return torch.any(masks, dim=0)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import os
from collections.abc import Iterable
from pathlib import Path

import scipy.sparse as sp
import torch

from anemoi.graphs.create import GraphCreator


class GraphExporter:
    """Class for exporting the graph to a sparse format."""

    def __init__(
        self,
        graph: str | Path,
        output_path: str | Path,
        edges_name: Iterable[tuple[str, str, str]] = None,
        edge_attribute_name: str = None,
        **kwargs,
    ):
        if isinstance(graph, Path) or isinstance(graph, str):
            if graph.endswith(".pt"):
                self.graph = torch.load(graph, weights_only=False, map_location="cpu")
            elif not graph.endswith(".yaml"):
                raise ValueError("The argument graph must be an actual graph (.pt) or a recipe to build one (.yaml).")

        self.graph = GraphCreator(graph).create(save_path=None).to("cpu")

        self.edges_name = self.graph.edge_types if edges_name is None else edges_name
        self.edge_attribute_name = edge_attribute_name
        self.output_path = output_path

    def get_edges_info(self, source_name: str, target_name: str) -> tuple[torch.Tensor, torch.Tensor]:
        """Get the edges info from the graph."""
        edge_index = self.graph[(source_name, "to", target_name)]["edge_index"]

        if self.edge_attribute_name is None:
            edge_attribute = torch.ones(edge_index.shape[1])
        else:
            assert (
                self.edge_attribute_name in self.graph[(source_name, "to", target_name)]
            ), f"Edge attribute {self.edge_attribute_name} not found in graph edges from {source_name} to {target_name}"
            edge_attribute = self.graph[(source_name, "to", target_name)][self.edge_attribute_name].squeeze()

        assert edge_attribute.ndim == 1, f"Edge attribute {self.edge_attribute_name} should be 1D."
        return edge_index, edge_attribute

    def get_nodes_info(self, source_name: str, target_name: str) -> tuple[int, int]:
        """Get the nodes info from the graph."""
        num_source_nodes = self.graph[source_name].num_nodes
        num_target_nodes = self.graph[target_name].num_nodes
        return num_source_nodes, num_target_nodes

    @staticmethod
    def get_sparse_matrix(edge_index, edge_attribute, num_source_nodes, num_target_nodes):
        """Create sparse matrix for y = A x.

        x is defined on source nodes, and output is on target nodes.

        Arguments
        ---------
        edge_index : torch.Tensor
            (2, E) tensor with rows: target nodes, cols: source nodes
        edge_attribute : torch.Tensor
            Edge weights/attributes
        num_source_nodes : int
            Number of source nodes (n_in)
        num_target_nodes : int
            Number of target nodes (n_out)

        Returns
        -------
        torch.sparse_coo_tensor
            Sparse COO tensor on target nodes
        """
        rows = edge_index[1]
        cols = edge_index[0]
        indices = torch.stack([rows, cols])

        A = torch.sparse_coo_tensor(
            indices,
            edge_attribute,
            (num_target_nodes, num_source_nodes),
            device=edge_index.device,
        )
        return A.coalesce()

    @staticmethod
    def convert_to_scipy_sparse(A):
        """Convert PyTorch sparse tensor to SciPy sparse matrix and save.

        Arguments
        ---------
        A : torch.sparse_coo_tensor
            Sparse tensor
        """
        # Get indices and values from PyTorch sparse tensor
        indices = A.indices().cpu().numpy()
        values = A.values().cpu().numpy()
        shape = A.shape

        # Create SciPy sparse matrix (COO format)
        scipy_sparse = sp.coo_matrix((values, (indices[0], indices[1])), shape=shape)
        return scipy_sparse

    def export(self):
        """Export the graph to a sparse format."""
        os.makedirs(self.output_path, exist_ok=True)
        for source_nodes, _, target_nodes in self.edges_name:
            edge_index, edge_attribute = self.get_edges_info(source_nodes, target_nodes)
            num_source_nodes, num_target_nodes = self.get_nodes_info(source_nodes, target_nodes)
            A = GraphExporter.get_sparse_matrix(edge_index, edge_attribute, num_source_nodes, num_target_nodes)
            A = GraphExporter.convert_to_scipy_sparse(A)

            output_path = Path(self.output_path) / f"{self.edge_attribute_name}-{source_nodes}_to_{target_nodes}.npz"
            sp.save_npz(output_path, A)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import pytest
import torch
from torch_geometric.data import HeteroData

from anemoi.graphs.processors.post_process import RemoveUnconnectedNodes
from anemoi.graphs.processors.post_process import RestrictEdgeLength
from anemoi.graphs.processors.post_process import SubsetNodesInArea


def test_remove_unconnected_nodes(graph_with_isolated_nodes: HeteroData):
    processor = RemoveUnconnectedNodes(nodes_name="test_nodes", ignore=None, save_mask_indices_to_attr=None)

    graph = processor.update_graph(graph_with_isolated_nodes)

    assert graph["test_nodes"].num_nodes == 4
    assert torch.equal(graph["test_nodes"].x, torch.tensor([[2], [3], [4], [5]]))
    assert "original_indices" not in graph["test_nodes"]


def test_remove_unconnected_nodes_with_indices_attr(graph_with_isolated_nodes: HeteroData):
    processor = RemoveUnconnectedNodes(
        nodes_name="test_nodes", ignore=None, save_mask_indices_to_attr="original_indices"
    )

    graph = processor.update_graph(graph_with_isolated_nodes)

    assert graph["test_nodes"].num_nodes == 4
    assert torch.equal(graph["test_nodes"].x, torch.tensor([[2], [3], [4], [5]]))
    assert torch.equal(graph["test_nodes", "to", "test_nodes"].edge_index, torch.tensor([[1, 2, 3], [0, 1, 2]]))
    assert torch.equal(graph["test_nodes"].original_indices, torch.tensor([[1], [2], [3], [4]]))


def test_remove_unconnected_nodes_with_ignore(graph_with_isolated_nodes: HeteroData):
    processor = RemoveUnconnectedNodes(nodes_name="test_nodes", ignore="mask_attr", save_mask_indices_to_attr=None)

    graph = processor.update_graph(graph_with_isolated_nodes)

    assert graph["test_nodes"].num_nodes == 5
    assert torch.equal(graph["test_nodes"].x, torch.tensor([[1], [2], [3], [4], [5]]))
    assert torch.equal(graph["test_nodes", "to", "test_nodes"].edge_index, torch.tensor([[2, 3, 4], [1, 2, 3]]))


@pytest.mark.parametrize(
    "nodes_name,ignore,save_mask_indices_to_attr",
    [
        ("test_nodes", None, "original_indices"),
        ("test_nodes", "mask_attr", None),
        ("test_nodes", None, None),
    ],
)
def test_remove_unconnected_nodes_parametrized(
    graph_with_isolated_nodes: HeteroData,
    nodes_name: str,
    ignore: str | None,
    save_mask_indices_to_attr: str | None,
):
    processor = RemoveUnconnectedNodes(
        nodes_name=nodes_name, ignore=ignore, save_mask_indices_to_attr=save_mask_indices_to_attr
    )

    graph = processor.update_graph(graph_with_isolated_nodes)

    assert isinstance(graph, HeteroData)
    pruned_nodes = 4 if ignore is None else 5
    assert graph[nodes_name].num_nodes == pruned_nodes

    if save_mask_indices_to_attr:
        assert save_mask_indices_to_attr in graph[nodes_name]
        assert graph[nodes_name][save_mask_indices_to_attr].ndim == 2
    else:
        assert graph[nodes_name].node_attrs() == graph_with_isolated_nodes[nodes_name].node_attrs()


def test_sort_edge_index_by_source_nodes(graph_nodes_and_edges: HeteroData):
    from anemoi.graphs.processors.post_process import SortEdgeIndexBySourceNodes

    processor = SortEdgeIndexBySourceNodes(descending=True)
    sorted_graph = processor.update_graph(graph_nodes_and_edges)

    expected_edge_index = torch.tensor([[3, 2, 1, 0], [2, 1, 0, 3]])

    sorted_edges = sorted_graph[("test_nodes", "to", "test_nodes")]
    assert torch.equal(sorted_edges.edge_index, expected_edge_index)
    assert torch.equal(sorted_edges.edge_attr, 10 * expected_edge_index[0][:, None])


def test_sort_edge_index_by_target_nodes(graph_nodes_and_edges: HeteroData):
    from anemoi.graphs.processors.post_process import SortEdgeIndexByTargetNodes

    processor = SortEdgeIndexByTargetNodes(descending=True)
    sorted_graph = processor.update_graph(graph_nodes_and_edges)

    expected_edge_index = torch.tensor([[0, 3, 2, 1], [3, 2, 1, 0]])

    sorted_edges = sorted_graph[("test_nodes", "to", "test_nodes")]
    assert torch.equal(sorted_edges.edge_index, expected_edge_index)
    assert torch.equal(sorted_edges.edge_attr, 10 * expected_edge_index[0][:, None])


def test_sort_edge_index_ascending_order(graph_nodes_and_edges: HeteroData):
    from anemoi.graphs.processors.post_process import SortEdgeIndexBySourceNodes

    processor = SortEdgeIndexBySourceNodes(descending=False)
    sorted_graph = processor.update_graph(graph_nodes_and_edges)

    expected_edge_index = torch.tensor([[0, 1, 2, 3], [3, 0, 1, 2]])

    sorted_edges = sorted_graph[("test_nodes", "to", "test_nodes")]
    assert torch.equal(sorted_edges.edge_index, expected_edge_index)
    assert torch.equal(sorted_edges.edge_attr, 10 * expected_edge_index[0][:, None])


def test_restrict_edge_length(graph_long_and_short_edges: HeteroData):
    """Test removal of all long ( > 1000km) edges."""
    graph = graph_long_and_short_edges
    expected_nodes_x = graph["test_nodes"].x

    short_mask = torch.tensor([1, 0, 0, 1], dtype=torch.bool)
    expected_edge_index = graph["test_nodes", "to", "test_nodes"].edge_index[:, short_mask]

    processor = RestrictEdgeLength("test_nodes", "test_nodes", 1000)
    restricted_graph = processor.update_graph(graph)

    assert torch.equal(restricted_graph["test_nodes", "to", "test_nodes"].edge_index, expected_edge_index)
    assert torch.equal(restricted_graph["test_nodes"].x, expected_nodes_x)


def test_restrict_edge_length_source_mask(graph_long_and_short_edges: HeteroData):
    """Test removal of all long ( > 1000km) edges with source in southern hemisphere."""
    graph = graph_long_and_short_edges
    expected_nodes_x = graph["test_nodes"].x

    long_southern_source_mask = torch.tensor([0, 1, 0, 0], dtype=torch.bool)
    expected_edge_index = graph["test_nodes", "to", "test_nodes"].edge_index[:, ~long_southern_source_mask]

    processor = RestrictEdgeLength("test_nodes", "test_nodes", 1000, source_mask_attr_name="southern_hemisphere_mask")
    restricted_graph = processor.update_graph(graph)

    assert torch.equal(restricted_graph["test_nodes", "to", "test_nodes"].edge_index, expected_edge_index)
    assert torch.equal(restricted_graph["test_nodes"].x, expected_nodes_x)


def test_restrict_edge_length_target_mask(graph_long_and_short_edges: HeteroData):
    """Test removal of all long ( > 1000km) edges with target in southern hemisphere."""
    graph = graph_long_and_short_edges
    expected_nodes_x = graph["test_nodes"].x

    long_southern_target_mask = torch.tensor([0, 0, 1, 0], dtype=torch.bool)
    expected_edge_index = graph["test_nodes", "to", "test_nodes"].edge_index[:, ~long_southern_target_mask]

    processor = RestrictEdgeLength("test_nodes", "test_nodes", 1000, target_mask_attr_name="southern_hemisphere_mask")
    restricted_graph = processor.update_graph(graph)

    assert torch.equal(restricted_graph["test_nodes", "to", "test_nodes"].edge_index, expected_edge_index)
    assert torch.equal(restricted_graph["test_nodes"].x, expected_nodes_x)


def test_subset_nodes_in_area(graph_long_and_short_edges: HeteroData):
    processor = SubsetNodesInArea("test_nodes", (90, -1, -90, 1))
    graph = processor.update_graph(graph_long_and_short_edges)

    # test the processor removes the nodes outside
    assert graph["test_nodes"].num_nodes == 2
    # test the processor removes the edges from/to removed nodes
    assert torch.all(graph["test_nodes", "to", "test_nodes"].edge_index == torch.tensor([[0], [1]]))



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from itertools import chain
from pathlib import Path

import torch
from hydra.utils import instantiate
from omegaconf import DictConfig
from torch_geometric.data import HeteroData

from anemoi.utils.config import DotDict

LOGGER = logging.getLogger(__name__)


class GraphCreator:
    """Graph creator."""

    config: DotDict

    def __init__(
        self,
        config: str | Path | DotDict | DictConfig,
    ):
        if isinstance(config, Path) or isinstance(config, str):
            self.config = DotDict.from_file(config)
        elif isinstance(config, DictConfig):
            self.config = DotDict(config)
        else:
            self.config = config

    def update_graph(self, graph: HeteroData) -> HeteroData:
        """Update the graph.

        It instantiates the node builders and edge builders defined in the configuration
        file and applies them to the graph.

        Parameters
        ----------
        graph : HeteroData
            The input graph to be updated.

        Returns
        -------
        HeteroData
            The updated graph with new nodes and edges added based on the configuration.
        """
        for nodes_name, nodes_cfg in self.config.get("nodes", {}).items():
            graph = instantiate(nodes_cfg.node_builder, name=nodes_name).update_graph(
                graph, attrs_config=nodes_cfg.get("attributes", {})
            )

        for edges_cfg in self.config.get("edges", {}):
            for edge_builder_cfg in edges_cfg.edge_builders:
                edge_builder = instantiate(
                    edge_builder_cfg,
                    source_name=edges_cfg.source_name,
                    target_name=edges_cfg.target_name,
                )
                graph = edge_builder.update_graph(graph, attrs_config=None)

            graph = edge_builder.register_attributes(graph, edges_cfg.get("attributes", {}))

        if graph.num_nodes == 0:
            LOGGER.warning("The graph that was created has no nodes. Please check your graph configuration file.")

        return graph

    def clean(self, graph: HeteroData) -> HeteroData:
        """Remove private attributes used during creation from the graph.

        Parameters
        ----------
        graph : HeteroData
            Generated graph

        Returns
        -------
        HeteroData
            Cleaned graph
        """
        LOGGER.info("Cleaning graph.")
        for type_name in chain(graph.node_types, graph.edge_types):
            attr_names_to_remove = [attr_name for attr_name in graph[type_name] if attr_name.startswith("_")]
            for attr_name in attr_names_to_remove:
                del graph[type_name][attr_name]
                LOGGER.info(f"{attr_name} deleted from graph.")

        return graph

    def post_process(self, graph: HeteroData) -> HeteroData:
        """Allow post-processing of the resulting graph.

        This method applies any configured post-processors to the graph,
        which can modify or enhance the graph structure or attributes.

        Parameters
        ----------
        graph : HeteroData
            The graph to be post-processed.

        Returns
        -------
        HeteroData
            The post-processed graph.

        Notes
        -----
        Post-processors are applied in the order they are specified in the configuration.
        Each post-processor should implement an `update_graph` method that takes and returns a HeteroData object.
        """
        for processor in self.config.get("post_processors", []):
            graph = instantiate(processor).update_graph(graph, graph_config=self.config)

        return graph

    def save(self, graph: HeteroData, save_path: Path, overwrite: bool = False) -> None:
        """Save the generated graph to the output path.

        Parameters
        ----------
        graph : HeteroData
            generated graph
        save_path : Path
            location to save the graph
        overwrite : bool, optional
            whether to overwrite existing graph file, by default False
        """
        save_path = Path(save_path)

        if not save_path.exists() or overwrite:
            save_path.parent.mkdir(parents=True, exist_ok=True)
            torch.save(graph, save_path)
            LOGGER.info(f"Graph saved at {save_path}.")
        else:
            # The error is only logged for compatibility with multi-gpu training in anemoi-training.
            # Currently, distributed graph creation is not supported so we create the same graph in each gpu.
            LOGGER.error(
                f"Graph not saved because {save_path} already exists. If this occurred during a multi-process or multi-GPU run, another process likely saved it first. If you intended to recreate it, rerun with overwrite=True."
            )

    def create(self, save_path: Path | None = None, overwrite: bool = False) -> HeteroData:
        """Create the graph and save it to the output path.

        Parameters
        ----------
        save_path : Path, optional
            location to save the graph, by default None
        overwrite : bool, optional
            whether to overwrite existing graph file, by default False

        Returns
        -------
        HeteroData
            created graph object
        """
        graph = HeteroData()
        graph = self.update_graph(graph)
        graph = self.clean(graph)
        graph = self.post_process(graph)

        if save_path is None:
            LOGGER.warning("No output path specified. The graph will not be saved.")
        else:
            self.save(graph, save_path, overwrite)

        return graph



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import numpy as np
import torch


def cartesian_to_latlon_rad(xyz: np.ndarray) -> np.ndarray:
    """3D to lat-lon (in radians) conversion.

    Convert 3D coordinates of points to its coordinates on the sphere containing
    them.

    Parameters
    ----------
    xyz : np.ndarray
        The 3D coordinates of points.

    Returns
    -------
    np.ndarray
        A 2D array of the coordinates of shape (N, 2) in radians.
    """
    lat = np.arcsin(xyz[..., 2] / (xyz**2).sum(axis=1))
    lon = np.arctan2(xyz[..., 1], xyz[..., 0])
    return np.array((lat, lon), dtype=np.float32).transpose()


def latlon_rad_to_cartesian_np(locations: np.ndarray, radius: float = 1) -> np.ndarray:
    """Convert planar coordinates to 3D coordinates in a sphere.

    Parameters
    ----------
    locations : np.ndarray of shape (N, 2)
        The 2D coordinates of the points, in radians.
    radius : float, optional
        The radius of the sphere containing los points. Defaults to the unit sphere.

    Returns
    -------
    np.ndarray of shape (N, 3)
        3D coordinates of the points in the sphere.
    """
    latr, lonr = locations[..., 0], locations[..., 1]
    x = radius * np.cos(latr) * np.cos(lonr)
    y = radius * np.cos(latr) * np.sin(lonr)
    z = radius * np.sin(latr)
    return np.stack((x, y, z), axis=-1)


def latlon_rad_to_cartesian(locations: torch.Tensor, radius: float = 1) -> torch.Tensor:
    """Convert planar coordinates to 3D coordinates in a sphere.

    Parameters
    ----------
    locations : torch.Tensor of shape (N, 2)
        The 2D coordinates of the points, in radians.
    radius : float, optional
        The radius of the sphere containing los points. Defaults to the unit sphere.

    Returns
    -------
    torch.Tensor of shape (N, 3)
        3D coordinates of the points in the sphere.
    """
    latr, lonr = locations[..., 0], locations[..., 1]
    x = radius * torch.cos(latr) * torch.cos(lonr)
    y = radius * torch.cos(latr) * torch.sin(lonr)
    z = radius * torch.sin(latr)
    return torch.stack((x, y, z), dim=-1)


def latlon_rad_to_sincos(locations: torch.Tensor) -> torch.Tensor:
    """Convert planar coordinates to sin-cos representation.

    Parameters
    ----------
    locations : torch.Tensor of shape (N, 2)
        The 2D coordinates of the points, in radians.

    Returns
    -------
    torch.Tensor of shape (N, 4)
        2D coordinates of the points in sin-cos representation.
    """
    return torch.cat([torch.sin(locations), torch.cos(locations)], dim=-1)


def sincos_to_latlon_rad(sincos_coords: torch.Tensor) -> torch.Tensor:
    """Convert sin-cos representation to planar coordinates.

    Parameters
    ----------
    sincos_coords : torch.Tensor of shape (N, 4)
        The 2D coordinates of the points in sin-cos representation.

    Returns
    -------
    torch.Tensor of shape (N, 2)
        The 2D coordinates of the points, in radians. Latitudes in [-pi/2, pi/2] and longitudes in [0, 2*pi].
    """
    ndim = sincos_coords.shape[1] // 2
    sin_values = sincos_coords[:, :ndim]
    cos_values = sincos_coords[:, ndim:]
    coords = torch.atan2(sin_values, cos_values)
    coords[:, 1] = torch.remainder(coords[:, 1], 2 * torch.pi)
    return coords



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest
from sklearn.neighbors import NearestNeighbors
from torch_geometric.data import HeteroData

from anemoi.graphs.generate.masks import KNNAreaMaskBuilder


def test_init():
    """Test KNNAreaMaskBuilder initialization."""
    mask_builder1 = KNNAreaMaskBuilder("nodes")
    mask_builder2 = KNNAreaMaskBuilder("nodes", margin_radius_km=120)
    mask_builder3 = KNNAreaMaskBuilder("nodes", mask_attr_name="mask")
    mask_builder4 = KNNAreaMaskBuilder("nodes", margin_radius_km=120, mask_attr_name="mask")

    assert isinstance(mask_builder1, KNNAreaMaskBuilder)
    assert isinstance(mask_builder2, KNNAreaMaskBuilder)
    assert isinstance(mask_builder3, KNNAreaMaskBuilder)
    assert isinstance(mask_builder4, KNNAreaMaskBuilder)

    assert isinstance(mask_builder1.nearest_neighbour, NearestNeighbors)
    assert isinstance(mask_builder2.nearest_neighbour, NearestNeighbors)
    assert isinstance(mask_builder3.nearest_neighbour, NearestNeighbors)
    assert isinstance(mask_builder4.nearest_neighbour, NearestNeighbors)


@pytest.mark.parametrize("margin", [-1, "120", None])
def test_fail_init_wrong_margin(margin: int):
    """Test KNNAreaMaskBuilder initialization with invalid margin."""
    with pytest.raises(AssertionError):
        KNNAreaMaskBuilder("nodes", margin_radius_km=margin)


@pytest.mark.parametrize("mask", [None, "mask"])
def test_fit(graph_with_nodes: HeteroData, mask: str):
    """Test KNNAreaMaskBuilder fit."""
    mask_builder = KNNAreaMaskBuilder("test_nodes", mask_attr_name=mask)
    assert not hasattr(mask_builder.nearest_neighbour, "n_samples_fit_")

    mask_builder.fit(graph_with_nodes)

    assert mask_builder.nearest_neighbour.n_samples_fit_ == graph_with_nodes["test_nodes"].num_nodes


def test_fit_fail(graph_with_nodes):
    """Test KNNAreaMaskBuilder fit with wrong graph."""
    mask_builder = KNNAreaMaskBuilder("wrong_nodes")
    with pytest.raises(AssertionError):
        mask_builder.fit(graph_with_nodes)



import pytest
from torch_geometric.data import HeteroData

from anemoi.graphs.nodes.builders.from_file import XArrayNodes


def test_update_graph_creates_expected_nodes(mock_zarr_dataset_file):
    graph = HeteroData()

    xarray_nodes = XArrayNodes(
        dataset=mock_zarr_dataset_file,
        name="xarray_nodes",
    )

    graph = xarray_nodes.update_graph(graph)

    assert "xarray_nodes" in graph.node_types
    assert graph["xarray_nodes"].num_nodes == 25


@pytest.mark.parametrize(
    "lat_key, lon_key",
    [
        ("invalid_latitude", "longitude"),
        ("latitude", "invalid_longitude"),
    ],
)
def test_throws_error_with_invalid_lat(mock_zarr_dataset_file, lat_key, lon_key):
    graph = HeteroData()

    xarray_nodes = XArrayNodes(
        dataset=mock_zarr_dataset_file,
        name="xarray_nodes",
        lat_key=lat_key,
        lon_key=lon_key,
    )

    with pytest.raises(AssertionError):
        graph = xarray_nodes.update_graph(graph)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import Iterable
from typing import Union

import torch
from hydra.utils import instantiate
from torch_geometric.data import HeteroData

from anemoi.graphs import EARTH_RADIUS
from anemoi.graphs.edges.attributes import EdgeLength
from anemoi.graphs.utils import NodesAxis
from anemoi.graphs.utils import get_edge_attributes

LOGGER = logging.getLogger(__name__)


class PostProcessor(ABC):
    """Base PostProcessor class."""

    @abstractmethod
    def update_graph(self, graph: HeteroData, **kwargs: Any) -> HeteroData:
        raise NotImplementedError(f"The {self.__class__.__name__} class does not implement the method update_graph().")


class BaseNodeMaskingProcessor(PostProcessor, ABC):
    """Base class for mask based node processor."""

    def __init__(
        self,
        nodes_name: Union[str, Iterable[str]],
        save_mask_indices_to_attr: str | None = None,
    ) -> None:
        self.nodes_names = (nodes_name,) if isinstance(nodes_name, str) else tuple(nodes_name)
        self.save_mask_indices_to_attr = save_mask_indices_to_attr
        super().__init__()

    def removing_nodes(self, graph: HeteroData, mask: torch.Tensor, nodes_name: str) -> HeteroData:
        """Remove nodes based on the mask passed."""
        for attr_name in graph[nodes_name].node_attrs():
            graph[nodes_name][attr_name] = graph[nodes_name][attr_name][mask]

        return graph

    def remove_edges(self, graph: HeteroData, mask: torch.Tensor, edges_name: tuple[str, str, str]) -> HeteroData:
        """Remove edges based on the mask passed."""
        for attr_name in graph[edges_name].edge_attrs():
            if attr_name == "edge_index":
                graph[edges_name][attr_name] = graph[edges_name][attr_name][:, mask]
            else:
                graph[edges_name][attr_name] = graph[edges_name][attr_name][mask]

        return graph

    def create_indices_mapper_from_mask(self, mask: torch.Tensor) -> dict[int, int]:
        return dict(zip(torch.where(mask)[0].tolist(), list(range(mask.sum()))))

    def update_edge_indices(self, graph: HeteroData, mask: torch.Tensor, nodes_name: str) -> HeteroData:
        """Update the edge indices to the new position of the nodes."""
        idx_mapping = self.create_indices_mapper_from_mask(mask)
        for edges_name in graph.edge_types:
            if edges_name[0] == nodes_name:
                valid_edges_mask = mask[graph[edges_name].edge_index[0].cpu()]
                if valid_edges_mask.any():
                    graph = self.remove_edges(graph, valid_edges_mask, edges_name)
                graph[edges_name].edge_index[0] = graph[edges_name].edge_index[0].cpu().apply_(idx_mapping.get)

            if edges_name[2] == nodes_name:
                valid_edges_mask = mask[graph[edges_name].edge_index[1].cpu()]
                if valid_edges_mask.any():
                    graph = self.remove_edges(graph, valid_edges_mask, edges_name)
                graph[edges_name].edge_index[1] = graph[edges_name].edge_index[1].cpu().apply_(idx_mapping.get)

        return graph

    @abstractmethod
    def compute_mask(self, graph: HeteroData, nodes_name: str) -> torch.Tensor: ...

    def add_attribute(self, graph: HeteroData, mask: torch.Tensor, nodes_name: str) -> HeteroData:
        """Add an attribute of the mask indices as node attribute."""
        if self.save_mask_indices_to_attr is not None:
            LOGGER.info(
                f"An attribute {self.save_mask_indices_to_attr} has been added with the indices to mask the nodes from the original graph."
            )
            mask_indices = torch.where(mask)[0].reshape((graph[nodes_name].num_nodes, -1))
            graph[nodes_name][self.save_mask_indices_to_attr] = mask_indices

        return graph

    def update_graph(self, graph: HeteroData, **kwargs: Any) -> HeteroData:
        """Post-process the graph.

        Parameters
        ----------
        graph: HeteroData
            The graph to post-process.
        kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        HeteroData
            The post-processed graph.
        """
        for nodes_name in self.nodes_names:
            mask = self.compute_mask(graph, nodes_name).cpu()
            LOGGER.info(f"Removing {(~mask).sum()} nodes from {nodes_name}.")
            graph = self.removing_nodes(graph, mask, nodes_name)
            graph = self.update_edge_indices(graph, mask, nodes_name)
            graph = self.add_attribute(graph, mask, nodes_name)
        return graph


class RemoveUnconnectedNodes(BaseNodeMaskingProcessor):
    """Remove unconnected nodes in the graph.

    Attributes
    ----------
    nodes_name: str | Iterable[str]
        Name of the unconnected nodes to remove.
    ignore: str, optional
        Name of an attribute to ignore when removing nodes. Nodes with
        this attribute set to True will not be removed.
    save_mask_indices_to_attr: str, optional
        Name of the attribute to save the mask indices. If provided,
        the indices of the kept nodes will be saved in this attribute.

    Methods
    -------
    compute_mask(graph)
        Compute the mask of the connected nodes.
    prune_graph(graph, mask)
        Prune the nodes with the specified mask.
    add_attribute(graph, mask)
        Add an attribute of the mask indices as node attribute.
    update_graph(graph)
        Post-process the graph.
    """

    def __init__(
        self,
        nodes_name: Union[str, Iterable[str]],
        save_mask_indices_to_attr: str | None = None,
        ignore: str | None = None,
    ) -> None:
        super().__init__(nodes_name, save_mask_indices_to_attr)
        self.ignore = ignore

    def compute_mask(self, graph: HeteroData, nodes_name: str) -> torch.Tensor:
        """Compute the mask of connected nodes."""
        nodes = graph[nodes_name]
        connected_mask = torch.zeros(nodes.num_nodes, dtype=torch.bool)

        if self.ignore is not None:
            LOGGER.info(f"The nodes with {self.ignore}=True will not be removed.")
            connected_mask[nodes[self.ignore].bool().squeeze()] = True

        for (source_name, _, target_name), edges in graph.edge_items():
            if source_name == nodes_name:
                connected_mask[edges.edge_index[0]] = True

            if target_name == nodes_name:
                connected_mask[edges.edge_index[1]] = True

        return connected_mask


class SubsetNodesInArea(BaseNodeMaskingProcessor):
    """Remove unconnected nodes in the graph.

    Attributes
    ----------
    nodes_name : str | Iterable[str]
        Name of the unconnected nodes to remove.
    area : tuple[float, float, float, float]
        Area of interest to crop the nodes, (north, west, south, east).
    save_mask_indices_to_attr : str, optional
        Name of the attribute to save the mask indices. If provided,
        the indices of the kept nodes will be saved in this attribute.

    Methods
    -------
    compute_mask(graph)
        Compute the mask of the connected nodes.
    prune_graph(graph, mask)
        Prune the nodes with the specified mask.
    add_attribute(graph, mask)
        Add an attribute of the mask indices as node attribute.
    update_graph(graph)
        Post-process the graph.
    """

    def __init__(
        self,
        nodes_name: Union[str, Iterable[str]],
        area: tuple[float, float, float, float],
        save_mask_indices_to_attr: str | None = None,
    ) -> None:
        super().__init__(nodes_name, save_mask_indices_to_attr)
        self.north, self.west, self.south, self.east = area

        # Convert area limits to [-90, 90]x[0,360]
        self.west, self.east = self.west % 360, (self.east - 1) % 360 + 1
        assert (self.west - self.east) % 360 != 0, "West and East limits must be different."

    def points_inside_area(
        self,
        lats: torch.Tensor,
        lons: torch.Tensor,
    ) -> torch.Tensor:
        # Convert lat & lons to [-90, 90]x[0,360)
        lons %= 360

        assert (-90 <= lats).all() and (lats <= 90).all(), "Latitude values must be in the range [-90, 90]."
        assert (0 <= lons).all() and (lons <= 360).all(), "Longitude values must be in the range [0, 360]."

        in_lat_range = (lats >= self.south) & (lats <= self.north)
        if self.west < self.east:
            in_lon_range = (lons >= self.west) & (lons <= self.east)
        else:
            # e.g: W=-10,E=10. They will be mapped to west=350, east=10.
            in_lon_range = (lons >= self.west) | (lons <= self.east)
        return in_lat_range & in_lon_range

    def compute_mask(self, graph: HeteroData, nodes_name: str) -> torch.Tensor:
        """Compute the mask of connected nodes."""
        latlons = torch.rad2deg(graph[nodes_name].x)
        mask = self.points_inside_area(lats=latlons[:, 0], lons=latlons[:, 1])
        assert mask.any(), f"The area selected does not contain any {nodes_name} node."
        return mask


class BaseSortEdgeIndex(PostProcessor, ABC):
    """Base class for sort edge indices processor."""

    nodes_axis: NodesAxis | None = None

    def __init__(self, descending: bool = True) -> None:
        assert self.nodes_axis is not None, f"{self.__class__.__name__} must define the nodes_axis class attribute."
        self.descending = descending
        super().__init__()

    def get_sorting_mask(self, edges: dict) -> torch.Tensor:
        sort_indices = torch.sort(edges["edge_index"], descending=self.descending, dim=1)
        return sort_indices.indices[self.nodes_axis.value]

    @staticmethod
    def get_edge_dim(edge_attr: str) -> int:
        # edge_index has shape (2, n_edges) and other edge attributes have (n_edges, attr_dim)
        return int(edge_attr == "edge_index")

    @staticmethod
    def sort_by_indices(x: torch.Tensor, indices: torch.Tensor, dim: int = 1) -> torch.Tensor:
        return x.index_select(dim=dim, index=indices)

    def update_graph(self, graph: HeteroData) -> HeteroData:
        """Sort all edge indices in the graph.

        Parameters
        ----------
        graph: HeteroData
            The graph to post-process.

        Returns
        -------
        HeteroData
            The post-processed graph.
        """
        for (src, to, dst), edges in graph.edge_items():
            sort_indices = self.get_sorting_mask(edges)
            for edge_attr_name in edges.edge_attrs():
                dim = BaseSortEdgeIndex.get_edge_dim(edge_attr_name)
                edge_attr = BaseSortEdgeIndex.sort_by_indices(edges[edge_attr_name], sort_indices, dim=dim)
                graph[(src, to, dst)][edge_attr_name] = edge_attr
        return graph


class SortEdgeIndexBySourceNodes(BaseSortEdgeIndex):
    nodes_axis = NodesAxis.SOURCE


class SortEdgeIndexByTargetNodes(BaseSortEdgeIndex):
    nodes_axis = NodesAxis.TARGET


class BaseEdgeMaskingProcessor(PostProcessor, ABC):
    """Base class for mask based edge processor.

    Attributes
    ----------
    source_name: str
        Name of the source nodes of edges to remove.
    target_name: str
        Name of the target nodes of edges to remove.
    """

    def __init__(
        self,
        source_name: str,
        target_name: str,
    ) -> None:
        self.source_name = source_name
        self.target_name = target_name
        self.edges_name = (self.source_name, "to", self.target_name)
        super().__init__()

    def removing_edges(self, graph: HeteroData, mask: torch.Tensor) -> HeteroData:
        """Remove edges based on the mask passed."""
        for attr_name in graph[self.edges_name].edge_attrs():
            if attr_name == "edge_index":
                graph[self.edges_name][attr_name] = graph[self.edges_name][attr_name].cpu()[:, mask]
            else:
                graph[self.edges_name][attr_name] = graph[self.edges_name][attr_name].cpu()[mask, :]

        return graph

    @abstractmethod
    def compute_mask(self, graph: HeteroData) -> torch.Tensor: ...

    def recompute_attributes(self, graph: HeteroData, graph_config: dict) -> HeteroData:
        """Recompute attributes"""
        edge_attributes = get_edge_attributes(graph_config, self.source_name, self.target_name)
        for attr_name, edge_attr_builder in edge_attributes.items():
            LOGGER.info(f"Recomputing edge attribute {attr_name}.")
            graph[self.edges_name][attr_name] = instantiate(edge_attr_builder)(
                x=(graph[self.source_name], graph[self.target_name]), edge_index=graph[self.edges_name].edge_index
            )
        return graph

    def update_graph(self, graph: HeteroData, **kwargs: Any) -> HeteroData:
        """Post-process the graph.

        Parameters
        ----------
        graph: HeteroData
            The graph to post-process.
        kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        HeteroData
            The post-processed graph.
        """
        mask = self.compute_mask(graph).cpu()
        LOGGER.info(f"Removing {(~mask).sum()} edges from {self.edges_name}.")
        graph = self.removing_edges(graph, mask)
        graph_config = kwargs.get("graph_config", {})
        graph = self.recompute_attributes(graph, graph_config)
        return graph


class RestrictEdgeLength(BaseEdgeMaskingProcessor):
    """Remove edges longer than a given treshold from the graph.

    Attributes
    ----------
    source_name: str
        Name of the source nodes of edges to remove.
    target_name: str
        Name of the target nodes of edges to remove.
    max_length_km: float
        The maximal length of edges not to be removed.
    source_mask_attr_name: str , optional
         the postprocessing will be restricted to edges with source node having True in this mask_attr
    target_mask_attr_name: str, optional
        the postprocessing will be restricted to edges with target node having True in this mask_attr
    Methods
    -------
    compute_mask(graph)
        Compute the mask of the relevant edges longer than max_length_km.
    """

    def __init__(
        self,
        source_name: str,
        target_name: str,
        max_length_km: float,
        source_mask_attr_name: str | None = None,
        target_mask_attr_name: str | None = None,
    ) -> None:
        super().__init__(source_name, target_name)
        self.treshold = max_length_km
        self.source_mask_attr_name = source_mask_attr_name
        self.target_mask_attr_name = target_mask_attr_name

    def compute_mask(self, graph: HeteroData) -> torch.Tensor:
        source_nodes = graph[self.source_name]
        target_nodes = graph[self.target_name]
        edge_index = graph[self.edges_name].edge_index
        lengths = EARTH_RADIUS * EdgeLength()(x=(source_nodes, target_nodes), edge_index=edge_index)
        mask = torch.where(lengths > self.treshold, False, True).squeeze().cpu()
        cases = [
            (self.source_mask_attr_name, source_nodes, 0),
            (self.target_mask_attr_name, target_nodes, 1),
        ]
        for mask_attr_name, nodes, i in cases:
            if mask_attr_name:
                attr_mask = nodes[mask_attr_name].squeeze()
                edge_mask = attr_mask[edge_index[i]].cpu()
                mask = torch.logical_or(mask, ~edge_mask)
        return mask



import numpy as np
import torch

from anemoi.graphs.edges.directional import compute_directions

tol = 1e-7


def test_compute_directions_unit_vectors():
    """Test that compute_directions always returns unit vectors."""
    # Test with various lat/lon pairs
    source = torch.tensor([[0.0, 0.0], [0.5, 0.5], [-0.3, 0.8], [np.pi / 4, 0.0]], dtype=torch.float32)
    target = torch.tensor([[0.5, 0.5], [-0.3, 0.8], [0.0, 0.0], [np.pi / 3, np.pi / 6]], dtype=torch.float32)

    directions = compute_directions(source, target)

    # Check shape
    assert directions.shape == (4, 2)

    # Check all are unit vectors
    norms = torch.norm(directions, dim=-1)
    assert torch.allclose(norms, torch.ones_like(norms), atol=1e-5)


def test_compute_directions_equator_to_pole():
    """Test direction from equator to north pole."""
    source = torch.tensor([[0.0, 0.0]], dtype=torch.float32)  # Equator at prime meridian
    target = torch.tensor([[np.pi / 2, 0.0]], dtype=torch.float32)  # North pole

    directions = compute_directions(source, target)

    # Should be a unit vector
    assert torch.allclose(torch.norm(directions), torch.tensor(1.0), atol=1e-5)

    # The exact direction depends on the rotation, but it should be well-defined
    assert directions.shape == (1, 2)
    assert torch.all(torch.isfinite(directions))


def test_compute_directions_equator_eastward():
    """Test direction along equator moving east."""
    source = torch.tensor([[0.0, 0.0]], dtype=torch.float32)  # Equator at 0°
    target = torch.tensor([[0.0, np.pi / 4]], dtype=torch.float32)  # Equator at 45°E

    directions = compute_directions(source, target)

    # Should be a unit vector
    assert torch.allclose(torch.norm(directions), torch.tensor(1.0), atol=1e-5)

    # Direction should have reasonable components
    assert directions.shape == (1, 2)
    assert torch.all(torch.isfinite(directions))


def test_compute_directions_same_meridian():
    """Test direction along same meridian (longitude) moving north."""
    source = torch.tensor([[np.pi / 6, 0.0], [np.pi / 6, np.pi / 4]], dtype=torch.float32)  # 30°N at 0° and 45°E
    target = torch.tensor([[np.pi / 3, 0.0], [np.pi / 3, np.pi / 4]], dtype=torch.float32)  # 60°N at 0° and 45°E

    directions = compute_directions(source, target)

    # Should be unit vectors
    norms = torch.norm(directions, dim=-1)
    assert torch.allclose(norms, torch.ones_like(norms), atol=1e-5)

    # Directions should be valid and finite
    assert torch.all(torch.isfinite(directions))
    assert directions.shape == (2, 2)


def test_compute_directions_random_points():
    """Test with random points to ensure robustness."""
    torch.manual_seed(42)

    # Generate 100 random lat/lon pairs
    n = 100
    source_lat = (torch.rand(n) - 0.5) * np.pi  # -90° to 90°
    source_lon = (torch.rand(n) - 0.5) * 2 * np.pi  # -180° to 180°
    target_lat = (torch.rand(n) - 0.5) * np.pi
    target_lon = (torch.rand(n) - 0.5) * 2 * np.pi

    source = torch.stack([source_lat, source_lon], dim=-1)
    target = torch.stack([target_lat, target_lon], dim=-1)

    directions = compute_directions(source, target)

    # Check shape
    assert directions.shape == (n, 2)

    # All should be unit vectors
    norms = torch.norm(directions, dim=-1)
    assert torch.allclose(norms, torch.ones_like(norms), atol=1e-4)

    # All should be finite
    assert torch.all(torch.isfinite(directions))


def test_compute_directions_nearby_points():
    """Test with nearby points (edge case for numerical stability)."""
    source = torch.tensor([[0.0, 0.0]], dtype=torch.float32)
    target = torch.tensor([[0.001, 0.001]], dtype=torch.float32)  # Very close

    directions = compute_directions(source, target)

    # Should still produce a valid unit vector
    assert torch.allclose(torch.norm(directions), torch.tensor(1.0), atol=1e-5)
    assert torch.all(torch.isfinite(directions))


def test_compute_directions_antipodal_points():
    """Test with antipodal points (opposite sides of sphere)."""
    source = torch.tensor([[np.pi / 2, 0.0]], dtype=torch.float32)  # North pole
    target = torch.tensor([[-np.pi / 2, 0.0]], dtype=torch.float32)  # South pole

    directions = compute_directions(source, target)

    # Should produce a valid unit vector even for this extreme case
    assert torch.allclose(torch.norm(directions), torch.tensor(1.0), atol=1e-5)
    assert torch.all(torch.isfinite(directions))


def test_compute_directions_consistency():
    """Test that the function produces consistent results."""
    source = torch.tensor([[0.0, 0.0], [0.0, 0.0]], dtype=torch.float32)
    target = torch.tensor([[0.5, 0.5], [0.5, 0.5]], dtype=torch.float32)

    directions = compute_directions(source, target)

    # Same inputs should produce same outputs
    assert torch.allclose(directions[0], directions[1], atol=1e-6)


def test_compute_directions_known_values():
    """Test against known geometric cases with expected directions.

    The function rotates the coordinate system so target is at north pole (0,0,1),
    applies the same rotation to source, then returns the normalized (x,y) components
    of the rotated source on the tangent plane.
    """
    # Case 1: source=equator (1,0,0), target=north pole (0,0,1)
    # Target already at north pole, no rotation needed
    # Source stays at (1,0,0), so (x,y) = (1,0)
    source = torch.tensor([[0.0, 0.0]], dtype=torch.float32)
    target = torch.tensor([[np.pi / 2, 0.0]], dtype=torch.float32)

    directions = compute_directions(source, target)
    expected = torch.tensor([[1.0, 0.0]], dtype=torch.float32)

    assert torch.allclose(directions, expected, atol=1e-4), f"Expected {expected}, got {directions}"

    # Case 2: source=equator (1,0,0), target=30°N on prime meridian
    # Target at (cos30°, 0, sin30°) = (0.866, 0, 0.5)
    # Rotation around y-axis by 60° to move target to north pole
    # Source (1,0,0) rotated -> (0.5, 0, 0.866), normalized xy = (1, 0)
    source = torch.tensor([[0.0, 0.0]], dtype=torch.float32)
    target = torch.tensor([[np.pi / 6, 0.0]], dtype=torch.float32)  # 30°N

    directions = compute_directions(source, target)
    expected = torch.tensor([[1.0, 0.0]], dtype=torch.float32)

    assert torch.allclose(directions, expected, atol=1e-4), f"Expected {expected}, got {directions}"

    # Case 3: source=north pole (0,0,1), target=equator (1,0,0)
    # Rotation around y-axis by 90° to move target to north pole
    # Source (0,0,1) rotates to (-1,0,0), so (x,y) = (-1, 0)
    source = torch.tensor([[np.pi / 2, 0.0]], dtype=torch.float32)
    target = torch.tensor([[0.0, 0.0]], dtype=torch.float32)

    directions = compute_directions(source, target)
    expected = torch.tensor([[-1.0, 0.0]], dtype=torch.float32)

    assert torch.allclose(directions, expected, atol=1e-4), f"Expected {expected}, got {directions}"


def test_compute_directions_equator_90_degrees():
    """Test direction between points on equator 90° apart in longitude.

    For two points on the equator separated by 90° in longitude, after
    rotating the target to the north pole, the source position is determined
    by the rotation.
    """
    # Two points on equator, 90° apart in longitude
    source = torch.tensor([[0.0, 0.0]], dtype=torch.float32)  # (0°, 0°) -> (1, 0, 0) in 3D
    target = torch.tensor([[0.0, np.pi / 2]], dtype=torch.float32)  # (0°, 90°E) -> (0, 1, 0) in 3D

    directions = compute_directions(source, target)

    # Rotation around x-axis by 90° to move target (0,1,0) to north pole (0,0,1)
    # Source (1,0,0) is on the rotation axis, so it stays at (1,0,0)
    # (x, y) = (1, 0)
    expected = torch.tensor([[1.0, 0.0]], dtype=torch.float32)

    assert torch.allclose(directions, expected, atol=1e-4), f"Expected {expected}, got {directions}"


def test_compute_directions_small_latitude_steps():
    """Test that small steps in latitude produce consistent directions.

    For small steps along the same meridian (prime meridian, lon=0),
    the target is rotated to north pole around the y-axis, and source
    follows the same rotation. The (x,y) projection should be consistent.
    """
    # Multiple small steps along prime meridian
    latitudes = torch.tensor([0.0, 0.1, 0.2, 0.3], dtype=torch.float32)

    for i in range(len(latitudes) - 1):
        source = torch.tensor([[latitudes[i], 0.0]], dtype=torch.float32)
        target = torch.tensor([[latitudes[i + 1], 0.0]], dtype=torch.float32)

        directions = compute_directions(source, target)

        # For points on prime meridian, rotation is around y-axis
        # Source (south of target) rotates to have positive x component
        # Expected direction is approximately (1, 0)
        expected = torch.tensor([[1.0, 0.0]], dtype=torch.float32)

        assert torch.allclose(
            directions, expected, atol=0.05
        ), f"Step {i} to {i+1}: Expected ~{expected}, got {directions}"


def test_compute_directions_symmetry():
    """Test rotational symmetry around poles.

    Points at the same latitude but rotated in longitude should produce
    directions that are rotated by the same angle.
    """
    # Test point at 45°N on prime meridian to north pole
    source1 = torch.tensor([[np.pi / 4, 0.0]], dtype=torch.float32)
    target1 = torch.tensor([[np.pi / 2, 0.0]], dtype=torch.float32)

    # Same test but rotated 90° in longitude
    source2 = torch.tensor([[np.pi / 4, np.pi / 2]], dtype=torch.float32)
    target2 = torch.tensor([[np.pi / 2, np.pi / 2]], dtype=torch.float32)

    dir1 = compute_directions(source1, target1)
    dir2 = compute_directions(source2, target2)

    # Both should have same magnitude
    assert torch.allclose(torch.norm(dir1), torch.norm(dir2), atol=1e-5)

    # The directions might differ due to different tangent plane orientations,
    # but both should be valid unit vectors
    assert torch.allclose(torch.norm(dir1), torch.tensor(1.0), atol=1e-5)
    assert torch.allclose(torch.norm(dir2), torch.tensor(1.0), atol=1e-5)


def test_compute_directions_pole_cases():
    """Test explicit handling of north and south pole cases."""
    # Case 1: Target at north pole (should use north pole case)
    source = torch.tensor([[0.0, 0.0], [np.pi / 4, np.pi / 2]], dtype=torch.float32)
    target = torch.tensor([[np.pi / 2, 0.0], [np.pi / 2, 0.0]], dtype=torch.float32)

    directions = compute_directions(source, target)

    # Should produce valid unit vectors
    assert torch.allclose(torch.norm(directions, dim=-1), torch.ones(2), atol=1e-5)
    assert torch.all(torch.isfinite(directions))

    # Case 2: Target at south pole (should use south pole case)
    source_south = torch.tensor([[0.0, 0.0], [np.pi / 4, np.pi / 2]], dtype=torch.float32)
    target_south = torch.tensor([[-np.pi / 2, 0.0], [-np.pi / 2, 0.0]], dtype=torch.float32)

    directions_south = compute_directions(source_south, target_south)

    # Should produce valid unit vectors
    assert torch.allclose(torch.norm(directions_south, dim=-1), torch.ones(2), atol=1e-5)
    assert torch.all(torch.isfinite(directions_south))

    # Case 3: Mixed batch (north pole, south pole, and normal)
    source_mixed = torch.tensor([[0.0, 0.0], [0.0, 0.0], [0.0, 0.0]], dtype=torch.float32)
    target_mixed = torch.tensor([[np.pi / 2, 0.0], [-np.pi / 2, 0.0], [np.pi / 4, 0.0]], dtype=torch.float32)

    directions_mixed = compute_directions(source_mixed, target_mixed)

    # All should produce valid unit vectors
    assert torch.allclose(torch.norm(directions_mixed, dim=-1), torch.ones(3), atol=1e-5)
    assert torch.all(torch.isfinite(directions_mixed))



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import itertools
import logging
import uuid
from functools import cached_property

import netCDF4
import numpy as np
import scipy
from typeguard import typechecked
from typing_extensions import Self

from anemoi.graphs.generate.utils import convert_adjacency_matrix_to_list
from anemoi.graphs.generate.utils import convert_list_to_adjacency_matrix
from anemoi.graphs.generate.utils import selection_matrix

LOGGER = logging.getLogger(__name__)


@typechecked
class NodeSet:
    """Stores nodes on the unit sphere."""

    id_iter: int = itertools.count()  # unique ID for each object
    gc_vertices: np.ndarray  # geographical (lat/lon) coordinates [rad], shape [:,2]

    def __init__(self, lon: np.ndarray, lat: np.ndarray):
        self.gc_vertices = np.column_stack((lon, lat))
        self.id = uuid.uuid4()

    @property
    def num_vertices(self) -> int:
        return self.gc_vertices.shape[0]

    @cached_property
    def cc_vertices(self):
        """Cartesian coordinates [rad], shape [:,3]."""
        return self._gc_to_cartesian()

    def __getitem__(self, mask: np.ndarray) -> Self:
        """Returns a new NodeSet with vertices selected by mask."""
        selected_gc = self.gc_vertices[mask]
        return NodeSet(selected_gc[:, 0], selected_gc[:, 1])

    def __add__(self, other: Self) -> Self:
        """concatenates two node sets."""
        gc_vertices = np.concatenate((self.gc_vertices, other.gc_vertices))
        return NodeSet(gc_vertices[:, 0], gc_vertices[:, 1])

    def __eq__(self, other: Self) -> bool:
        """Compares two node sets."""
        return self.id == other.id

    def _gc_to_cartesian(self, radius: float = 1.0) -> np.ndarray:
        """Returns Cartesian coordinates of the node set, shape [:,3]."""
        xyz = (
            radius * np.cos(lat_rad := self.gc_vertices[:, 1]) * np.cos(lon_rad := self.gc_vertices[:, 0]),
            radius * np.cos(lat_rad) * np.sin(lon_rad),
            radius * np.sin(lat_rad),
        )
        return np.stack(xyz, axis=-1)


@typechecked
class ICONMultiMesh:
    """Reads vertices and topology from an ICON grid file; creates multi-mesh."""

    icon_grid_filename: str
    uuidOfHGrid: str
    reflvl_vertex: np.ndarray
    max_level: int
    nodeset: NodeSet  # set of ICON grid vertices

    def __init__(self, icon_grid_filename: str, max_level: int | None = None):
        self.grid_filename = icon_grid_filename

        # open file, representing the finest level
        LOGGER.debug(f"{type(self).__name__}: read coordinates from ICON grid file '{self.grid_filename}'")
        with netCDF4.Dataset(self.grid_filename, "r") as ncfile:
            # read vertex coordinates
            vlon = read_coordinate_array(ncfile, "vlon", "vertex")
            vlat = read_coordinate_array(ncfile, "vlat", "vertex")

            reflvl_vertex = ncfile.variables["refinement_level_v"][:]
            assert ncfile.variables["refinement_level_v"].dimensions == ("vertex",)

            self.nodeset = NodeSet(vlon, vlat)
            self.uuidOfHGrid = ncfile.uuidOfHGrid

        self.reflvl_vertex = reflvl_vertex
        self.max_level = max_level if max_level is not None else reflvl_vertex.max()

        # restrict edge-vertex list to multi_mesh level "max_level":
        if self.max_level < self.reflvl_vertex.max():
            self.nodeset = self.nodeset[self.reflvl_vertex <= self.max_level]

    @cached_property
    def _vertices(self) -> tuple[list[np.ndarray], np.ndarray]:
        edge_vertices_fine, cell_vertices_fine = self._read_vertices_data()

        edge_vertices, cell_vertices = self._get_hierarchy_of_icon_edge_graphs(edge_vertices_fine, cell_vertices_fine)

        if self.max_level < len(edge_vertices):
            edge_vertices, cell_vertices = self._restrict_multi_mesh_edges_and_cells(edge_vertices, cell_vertices)
        return edge_vertices, cell_vertices

    @property
    def edge_vertices(self) -> list[np.ndarray]:
        """Returns the multi-mesh edges as a list of arrays of vertex indices."""
        return self._vertices[0]

    @property
    def cell_vertices(self) -> np.ndarray:
        """Returns the multi-mesh cell_vertices."""
        return self._vertices[1]

    @property
    def multi_mesh_edges(self) -> np.ndarray:
        """Returns the multi-mesh edges as an arrays of vertex indices."""
        # concatenate edge-vertex lists (= edges of the multi-level mesh):
        return np.concatenate([edges for edges in self.edge_vertices], axis=0)

    def _read_vertices_data(self):
        LOGGER.debug(f"{type(self).__name__}: read vertices data from ICON grid file '{self.grid_filename}'")
        with netCDF4.Dataset(self.grid_filename, "r") as ncfile:

            edge_vertices_fine = np.asarray(ncfile.variables["edge_vertices"][:] - 1, dtype=np.int64).transpose()
            assert ncfile.variables["edge_vertices"].dimensions == ("nc", "edge")

            cell_vertices_fine = np.asarray(ncfile.variables["vertex_of_cell"][:] - 1, dtype=np.int64).transpose()
            assert ncfile.variables["vertex_of_cell"].dimensions == ("nv", "cell")

        return edge_vertices_fine, cell_vertices_fine

    def _restrict_multi_mesh_edges_and_cells(
        self,
        edge_vertices: list[np.ndarray],
        cell_vertices: np.ndarray,
    ) -> tuple[list[np.ndarray], np.ndarray]:
        """Creates a new mesh with only the vertices at the desired level."""
        vertex_mask = self.reflvl_vertex < self.max_level
        num_vertices = vertex_mask.shape[0]
        vertex_glb2loc = np.zeros(num_vertices, dtype=int)
        vertex_glb2loc[vertex_mask] = np.arange(vertex_mask.sum())
        return (
            [vertex_glb2loc[vertices] for vertices in edge_vertices[: self.max_level + 1]],
            # cell_vertices: preserve negative indices (incomplete cells)
            np.where(cell_vertices >= 0, vertex_glb2loc[cell_vertices], cell_vertices),
        )

    def _get_hierarchy_of_icon_edge_graphs(
        self,
        edge_vertices_fine: np.ndarray,
        cell_vertices_fine: np.ndarray,
    ) -> tuple[list[np.ndarray], np.ndarray]:
        """Returns a list of edge-vertex relations (coarsest to finest level)."""

        edge_vertices = [edge_vertices_fine]  # list of edge-vertex relations (coarsest to finest level).

        num_vertices = self.reflvl_vertex.shape[0]
        # edge-to-vertex adjacency matrix with 2 non-zero entries per row:
        edge2vertex_matrix = convert_list_to_adjacency_matrix(edge_vertices_fine, num_vertices)
        # cell-to-vertex adjacency matrix with 3 non-zero entries per row:
        cell2vertex_matrix = convert_list_to_adjacency_matrix(cell_vertices_fine, num_vertices)
        vertex2vertex_matrix = edge2vertex_matrix.transpose() * edge2vertex_matrix
        vertex2vertex_matrix.setdiag(np.ones(num_vertices))  # vertices are self-connected

        selected_vertex_coarse = scipy.sparse.diags(np.ones(num_vertices), dtype=bool)

        # coarsen edge-vertex list from level `ilevel -> ilevel - 1`:
        for ilevel in reversed(range(1, self.reflvl_vertex.max() + 1)):
            LOGGER.debug(f"  edges[{ilevel}] = {edge_vertices[0].shape[0] : >9}")

            # define edge selection matrix (selecting only edges of which have
            # exactly one coarse vertex):
            #
            # get a boolean mask, matching all edges where one of its vertices
            # has refinement level index `ilevel`:
            ref_level_mask = self.reflvl_vertex[edge_vertices[0]] == ilevel
            edges_coarse = np.logical_xor(ref_level_mask[:, 0], ref_level_mask[:, 1])  # = bisected coarse edges
            idx_edge2edge = np.argwhere(edges_coarse).flatten()
            selected_edges = selection_matrix(idx_edge2edge, edges_coarse.shape[0])

            # define vertex selection matrix selecting only vertices of
            # level `ilevel`:
            idx_v_fine = np.argwhere(self.reflvl_vertex == ilevel).flatten()
            selected_vertex_fine = selection_matrix(idx_v_fine, num_vertices)
            # define vertex selection matrix, selecting only vertices of
            # level < `ilevel`, by successively removing `s_fine` from an identity matrix.
            selected_vertex_coarse.data[0][idx_v_fine] = False

            # create an adjacency matrix which links each fine level
            # vertex to its two coarser neighbor vertices:
            vertex2vertex_fine2coarse = selected_vertex_fine * vertex2vertex_matrix * selected_vertex_coarse
            # remove rows that have only one non-zero entry
            # (corresponding to incomplete parent triangles in LAM grids):
            csum = vertex2vertex_fine2coarse * np.ones((vertex2vertex_fine2coarse.shape[1], 1))
            selected_vertex2vertex = selection_matrix(
                np.argwhere(csum == 2).flatten(), vertex2vertex_fine2coarse.shape[0]
            )
            vertex2vertex_fine2coarse = selected_vertex2vertex * vertex2vertex_fine2coarse

            # then construct the edges-to-parent-vertex adjacency matrix:
            parent_edge_vertices = selected_edges * edge2vertex_matrix * vertex2vertex_fine2coarse
            # again, we have might have selected edges within
            # `selected_edges` which are part of an incomplete parent edge
            # (LAM case). We filter these here:
            csum = parent_edge_vertices * np.ones((parent_edge_vertices.shape[1], 1))
            selected_edge2edge = selection_matrix(np.argwhere(csum == 2).flatten(), parent_edge_vertices.shape[0])
            parent_edge_vertices = selected_edge2edge * parent_edge_vertices

            # note: the edges-vertex adjacency matrix still has duplicate
            # rows, since two child edges have the same parent edge.
            edge_vertices_coarse = convert_adjacency_matrix_to_list(parent_edge_vertices, ncols_per_row=2)
            edge_vertices.insert(0, edge_vertices_coarse)

            # store cell-to-vert adjacency matrix
            if ilevel > self.max_level:
                cell2vertex_matrix = cell2vertex_matrix * vertex2vertex_fine2coarse
                # similar to the treatment above, we need to handle
                # coarse LAM cells which are incomplete.
                csum = cell2vertex_matrix * np.ones((cell2vertex_matrix.shape[1], 1))
                selected_cell2cell = selection_matrix(np.argwhere(csum == 3).flatten(), cell2vertex_matrix.shape[0])
                cell2vertex_matrix = selected_cell2cell * cell2vertex_matrix

            # replace edge-to-vertex and vert-to-vert adjacency matrices (for next level):
            if ilevel > 1:
                vertex2vertex_matrix = selected_vertex_coarse * vertex2vertex_matrix * vertex2vertex_fine2coarse
                edge2vertex_matrix = convert_list_to_adjacency_matrix(edge_vertices_coarse, num_vertices)

        # Fine-level cells outside of multi-mesh (LAM boundary)
        # correspond to empty rows in the adjacency matrix. We
        # substitute these by three additional, non-existent vertices:
        csum = 3 - cell2vertex_matrix * np.ones((cell2vertex_matrix.shape[1], 1))
        nvmax = cell2vertex_matrix.shape[1]
        cell2vertex_matrix = scipy.sparse.csr_matrix(scipy.sparse.hstack((cell2vertex_matrix, csum, csum, csum)))

        # build a list of cell-vertices [1..num_cells,1..3] for all
        # fine-level cells:
        cell_vertices = convert_adjacency_matrix_to_list(cell2vertex_matrix, remove_duplicates=False, ncols_per_row=3)

        # finally, translate non-existent vertices into "-1" indices:
        cell_vertices = np.where(
            cell_vertices >= nvmax,
            -np.ones(cell_vertices.shape, dtype=np.int32),
            cell_vertices,
        )

        return (edge_vertices, cell_vertices)


@typechecked
class ICONCellDataGrid:
    """Reads cell locations from an ICON grid file; builds grid-to-mesh edges based on ICON topology.

    This provides a bipartite graph, i.e. a graph defined on the NodeSets `self.nodeset` and an `ICONMultiMesh`
    """

    uuidOfHGrid: str
    nodeset: NodeSet  # set of ICON cell circumcenters
    max_level: int
    select_c: np.ndarray

    def __init__(self, icon_grid_filename: str, max_level: int | None = None):
        self.grid_filename = icon_grid_filename

        # open file, representing the finest level
        LOGGER.debug(f"{type(self).__name__}: read ICON grid file '{self.grid_filename}'")
        with netCDF4.Dataset(self.grid_filename, "r") as ncfile:
            # read cell circumcenter coordinates
            clon = read_coordinate_array(ncfile, "clon", "cell")
            clat = read_coordinate_array(ncfile, "clat", "cell")

            reflvl_cell = ncfile.variables["refinement_level_c"][:]
            assert ncfile.variables["refinement_level_c"].dimensions == ("cell",)

            self.uuidOfHGrid = ncfile.uuidOfHGrid

        self.max_level = max_level if max_level is not None else reflvl_cell.max()

        # restrict to level `max_level`:
        self.select_c = np.argwhere(reflvl_cell <= self.max_level)
        # generate source grid node set:
        self.nodeset = NodeSet(clon[self.select_c], clat[self.select_c])

    def get_grid2mesh_edges(self, multi_mesh: ICONMultiMesh) -> np.ndarray:
        """Create "grid-to-mesh" edges, ie. edges from (clat,clon) to the
        vertices of the multi-mesh.
        """
        num_cells = self.select_c.shape[0]
        num_vertices_per_cell = multi_mesh.cell_vertices.shape[1]
        src_list = np.kron(np.arange(num_cells), np.ones((1, num_vertices_per_cell), dtype=np.int64)).flatten()
        dst_list = multi_mesh.cell_vertices[self.select_c[:, 0], :].flatten()
        edge_vertices = np.stack((src_list, dst_list), axis=1, dtype=np.int64)
        return edge_vertices


# -------------------------------------------------------------


@typechecked
def read_coordinate_array(ncfile, arrname: str, dimname: str) -> np.ndarray:
    """Auxiliary routine, reading a coordinate array, checking consistency."""
    arr = ncfile.variables[arrname][:]
    assert ncfile.variables[arrname].dimensions == (dimname,)
    assert ncfile.variables[arrname].units == "radian"
    # netCDF4 returns all variables as numpy.ma.core.MaskedArray.
    # -> convert to regular arrays
    assert not arr.mask.any(), f"There are missing values in {arrname}"
    return arr.data



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest

from anemoi.graphs.edges import KNNEdges
from anemoi.graphs.edges import ReversedKNNEdges


@pytest.mark.parametrize("edge_builder", [KNNEdges, ReversedKNNEdges])
def test_init(edge_builder):
    """Test KNNEdges initialization."""
    edge_builder("test_nodes1", "test_nodes2", 3)


@pytest.mark.parametrize("edge_builder", [KNNEdges, ReversedKNNEdges])
@pytest.mark.parametrize("num_nearest_neighbours", [-1, 2.6, "hello", None])
def test_fail_init(edge_builder, num_nearest_neighbours: str):
    """Test KNNEdges initialization with invalid number of nearest neighbours."""
    with pytest.raises(AssertionError):
        edge_builder("test_nodes1", "test_nodes2", num_nearest_neighbours)


@pytest.mark.parametrize("edge_builder", [KNNEdges, ReversedKNNEdges])
def test_knn(edge_builder, graph_with_nodes):
    """Test KNNEdges."""
    builder = edge_builder("test_nodes", "test_nodes", 3)
    graph = builder.update_graph(graph_with_nodes)
    assert ("test_nodes", "to", "test_nodes") in graph.edge_types



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
import time
from abc import ABC
from abc import abstractmethod
from importlib.util import find_spec

import numpy as np
import torch
from hydra.utils import instantiate
from torch_geometric.data import HeteroData
from torch_geometric.data.storage import NodeStorage

from anemoi.graphs.edges.builders.masking import NodeMaskingMixin
from anemoi.graphs.utils import concat_edges
from anemoi.graphs.utils import get_distributed_device
from anemoi.utils.config import DotDict

LOGGER = logging.getLogger(__name__)

TORCH_CLUSTER_AVAILABLE = find_spec("torch_cluster") is not None

TORCH_CLUSTER_INSTRUCTIONS = r"""The 'torch-cluster' library is not installed.
Installing 'torch-cluster' can significantly improve performance for graph creation.
You can install it using:
    TORCH_VERSION=$(python -c "import torch; print(torch.__version__)")
    pip install torch-cluster -f https://data.pyg.org/whl/torch-${TORCH_VERSION}.html
"""


class BaseEdgeBuilder(ABC):
    """Base class for edge builders."""

    def __init__(
        self,
        source_name: str,
        target_name: str,
        source_mask_attr_name: str | None = None,
        target_mask_attr_name: str | None = None,
    ):
        self.source_name = source_name
        self.target_name = target_name
        self.source_mask_attr_name = source_mask_attr_name
        self.target_mask_attr_name = target_mask_attr_name
        self.device = get_distributed_device()

    @property
    def name(self) -> tuple[str, str, str]:
        """Name of the edge subgraph."""
        return self.source_name, "to", self.target_name

    @abstractmethod
    def compute_edge_index(self, source_nodes: NodeStorage, target_nodes: NodeStorage) -> torch.Tensor: ...

    def prepare_node_data(self, graph: HeteroData) -> tuple[NodeStorage, NodeStorage]:
        """Prepare node information and get source and target nodes."""
        return graph[self.source_name], graph[self.target_name]

    def get_edge_index(self, graph: HeteroData) -> torch.Tensor:
        """Get the edge index."""
        source_nodes, target_nodes = self.prepare_node_data(graph)
        edge_index = self.compute_edge_index(source_nodes, target_nodes)
        return edge_index.to(dtype=torch.int32, device=self.device)

    def register_edges(self, graph: HeteroData) -> HeteroData:
        """Register edges in the graph.

        Parameters
        ----------
        graph : HeteroData
            The graph to register the edges.

        Returns
        -------
        HeteroData
            The graph with the registered edges.
        """
        edge_index = self.get_edge_index(graph)
        edge_type = type(self).__name__

        if "edge_index" in graph[self.name]:
            # Expand current edge indices
            graph[self.name].edge_index = concat_edges(graph[self.name].edge_index, edge_index)
            if edge_type not in graph[self.name].edge_type:
                graph[self.name].edge_type = graph[self.name].edge_type + "," + edge_type
            return graph

        # Register new edge indices
        graph[self.name].edge_index = edge_index
        graph[self.name].edge_type = edge_type
        return graph

    def register_attributes(self, graph: HeteroData, config: DotDict) -> HeteroData:
        """Register attributes in the edges of the graph specified.

        Parameters
        ----------
        graph : HeteroData
            The graph to register the attributes.
        config : DotDict
            The configuration of the attributes.

        Returns
        -------
        HeteroData
            The graph with the registered attributes.
        """
        for attr_name, attr_config in config.items():
            edge_index = graph[self.name].edge_index
            edge_attribute_builder = instantiate(attr_config)
            graph[self.name][attr_name] = edge_attribute_builder(
                x=(graph[self.name[0]], graph[self.name[2]]), edge_index=edge_index
            )
        return graph

    def update_graph(self, graph: HeteroData, attrs_config: DotDict | None = None) -> HeteroData:
        """Update the graph with the edges.

        Parameters
        ----------
        graph : HeteroData
            The graph.
        attrs_config : DotDict
            The configuration of the edge attributes.

        Returns
        -------
        HeteroData
            The graph with the edges.
        """
        t0 = time.time()
        graph = self.register_edges(graph)
        t1 = time.time()
        LOGGER.debug("Time to register edge indices (%s): %.2f s", self.__class__.__name__, t1 - t0)

        if attrs_config is not None:
            t0 = time.time()
            graph = self.register_attributes(graph, attrs_config)
            t1 = time.time()
            LOGGER.debug("Time to register edge attribute (%s): %.2f s", self.__class__.__name__, t1 - t0)

        return graph


class BaseDistanceEdgeBuilders(BaseEdgeBuilder, NodeMaskingMixin, ABC):
    """Base class for edge builders based on distance."""

    @abstractmethod
    def _compute_edge_index_pyg(self, source_coords: NodeStorage, target_coords: NodeStorage) -> np.ndarray: ...

    @abstractmethod
    def _compute_adj_matrix_sklearn(self, source_coords: NodeStorage, target_coords: NodeStorage) -> np.ndarray: ...

    def compute_edge_index(self, source_nodes: NodeStorage, target_nodes: NodeStorage) -> torch.Tensor:
        """Compute the edge indices.

        Parameters
        ----------
        source_nodes : NodeStorage
            The source nodes.
        target_nodes : NodeStorage
            The target nodes.

        Returns
        -------
        torch.Tensor of shape (2, num_edges)
            Indices of source and target nodes connected by an edge.
        """
        source_coords, target_coords = self.get_cartesian_node_coordinates(source_nodes, target_nodes)

        if TORCH_CLUSTER_AVAILABLE:
            edge_index = self._compute_edge_index_pyg(source_coords, target_coords)
            edge_index = self.undo_masking_edge_index(edge_index, source_nodes, target_nodes)
        else:
            LOGGER.warning(TORCH_CLUSTER_INSTRUCTIONS)
            adj_matrix = self._compute_adj_matrix_sklearn(source_coords, target_coords)
            adj_matrix = self.undo_masking_adj_matrix(adj_matrix, source_nodes, target_nodes)
            edge_index = torch.from_numpy(np.stack([adj_matrix.col, adj_matrix.row], axis=0))

        return edge_index



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from enum import Enum

import torch
from sklearn.neighbors import NearestNeighbors

from anemoi.graphs.generate.transforms import latlon_rad_to_cartesian


def get_distributed_device() -> torch.device:
    """Get the distributed device.

    Returns
    -------
    torch.device
        The distributed device.
    """
    if torch.cuda.is_available():
        import os

        local_rank = int(os.environ.get("SLURM_LOCALID", 0))
        device = torch.device(f"cuda:{local_rank}")
    else:
        device = "cpu"

    return device


def get_nearest_neighbour(coords_rad: torch.Tensor, mask: torch.Tensor | None = None) -> NearestNeighbors:
    """Get NearestNeighbour object fitted to coordinates.

    Parameters
    ----------
    coords_rad : torch.Tensor
        corrdinates in radians
    mask : torch.Tensor, optional
        mask to remove nodes, by default None

    Returns
    -------
    NearestNeighbors
        fitted NearestNeighbour object
    """
    assert mask is None or mask.shape == (
        coords_rad.shape[0],
        1,
    ), "Mask must have the same shape as the number of nodes."

    nearest_neighbour = NearestNeighbors(metric="euclidean", n_jobs=4)

    nearest_neighbour.fit(coords_rad)

    return nearest_neighbour


def get_grid_reference_distance(coords_rad: torch.Tensor, mask: torch.Tensor | None = None) -> float:
    """Get the reference distance of the grid.

    It is the maximum distance of a node in the mesh with respect to its nearest neighbour.

    Parameters
    ----------
    coords_rad : torch.Tensor
        corrdinates in radians
    mask : torch.Tensor, optional
        mask to remove nodes, by default None

    Returns
    -------
    float
        The reference distance of the grid.
    """
    xyz = latlon_rad_to_cartesian(coords_rad)
    nearest_neighbours = get_nearest_neighbour(xyz, mask)
    dists, _ = nearest_neighbours.kneighbors(xyz, n_neighbors=2, return_distance=True)
    return dists[dists > 0].max()


def concat_edges(edge_indices1: torch.Tensor, edge_indices2: torch.Tensor) -> torch.Tensor:
    """Concat edges

    Parameters
    ----------
    edge_indices1: torch.Tensor
        Edge indices of the first set of edges. Shape: (2, num_edges1)
    edge_indices2: torch.Tensor
        Edge indices of the second set of edges. Shape: (2, num_edges2)

    Returns
    -------
    torch.Tensor
        Concatenated edge indices.
    """
    return torch.unique(torch.cat([edge_indices1, edge_indices2], axis=1), dim=1)


def haversine_distance(source_coords: torch.Tensor, target_coords: torch.Tensor) -> torch.Tensor:
    """Haversine distance.

    Parameters
    ----------
    source_coords : torch.Tensor of shape (N, 2)
        Source coordinates in radians.
    target_coords : torch.Tensor of shape (N, 2)
        Destination coordinates in radians.

    Returns
    -------
    torch.Tensor of shape (N,)
        Haversine distance between source and destination coordinates.
    """
    dlat = target_coords[:, 0] - source_coords[:, 0]
    dlon = target_coords[:, 1] - source_coords[:, 1]
    a = (
        torch.sin(dlat / 2) ** 2
        + torch.cos(source_coords[:, 0]) * torch.cos(target_coords[:, 0]) * torch.sin(dlon / 2) ** 2
    )
    c = 2 * torch.atan2(torch.sqrt(a), torch.sqrt(1 - a))
    return c


class NodesAxis(Enum):
    SOURCE = 0
    TARGET = 1


def get_edge_attributes(config: dict, source_name: str, target_name: str) -> dict:
    """Get edge attributes out of a graph config

    Parameters
    ----------
    config : dict
        The graph configuration.
    source_name : str
        Name of source nodes of edges to be considered
    target_name : str
        Name of target nodes of edges to be considered
    Returns
    -------
    dict
        Dictionary of the form {attribute_name: attribute}
    """
    attrs = {}
    for edges_config in config.get("edges", {}):
        if edges_config["source_name"] == source_name and edges_config["target_name"] == target_name:
            attrs.update(edges_config["attributes"])
    return attrs



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from anemoi.graphs.export import GraphExporter

from . import Command


class ExportToSparse(Command):
    """Export a graph edges to a sparse format.

    Example usage specifying an edge attribute:
    ```
    anemoi-graphs export-to-sparse graph.pt output_path --edge-attribute-name edge_attr
    ```

    Example usage specifying a subset of edges:
    ```
    anemoi-graphs export-to-sparse graph.pt output_path --edges-name data down --edges-name down data
    ```
    """

    internal = True
    timestamp = True

    def add_arguments(self, command_parser):
        command_parser.add_argument("graph", help="Path to the graph (a .PT file) or a config file defining the graph.")
        command_parser.add_argument("output_path", help="Path to store the inspection results.")
        command_parser.add_argument("--edge-attribute-name", default=None, help="Name of the edge attribute to export.")
        command_parser.add_argument(
            "--edges-name",
            nargs=2,
            action="append",
            metavar=("NODE1", "NODE2"),
            help="Specify an edge by its two node names. Can be used multiple times.",
        )

    def run(self, args):
        kwargs = vars(args)
        edges_name = kwargs.get("edges_name", None)
        if edges_name is not None:
            kwargs["edges_name"] = [(pair[0], "to", pair[1]) for pair in edges_name]

        GraphExporter(
            graph=kwargs["graph"],
            output_path=kwargs["output_path"],
            edge_attribute_name=kwargs["edge_attribute_name"],
            edges_name=kwargs["edges_name"],
        ).export()


command = ExportToSparse



# (C) Copyright 2024- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from .area_weights import AnemoiDatasetVariableWeights
from .area_weights import CosineLatWeightedAttribute
from .area_weights import IsolatitudeAreaWeights
from .area_weights import MaskedPlanarAreaWeights
from .area_weights import PlanarAreaWeights
from .area_weights import SphericalAreaWeights
from .area_weights import UniformWeights
from .boolean_op import BooleanAndMask
from .boolean_op import BooleanNot
from .boolean_op import BooleanOrMask
from .masks import CutOutMask
from .masks import GridsMask
from .masks import LimitedAreaMask
from .masks import NonmissingAnemoiDatasetVariable
from .masks import NonzeroAnemoiDatasetVariable

__all__ = [
    "GridsMask",
    "SphericalAreaWeights",
    "PlanarAreaWeights",
    "UniformWeights",
    "CutOutMask",
    "LimitedAreaMask",
    "MaskedPlanarAreaWeights",
    "NonmissingAnemoiDatasetVariable",
    "NonzeroAnemoiDatasetVariable",
    "BooleanAndMask",
    "BooleanNot",
    "BooleanOrMask",
    "CosineLatWeightedAttribute",
    "IsolatitudeAreaWeights",
    "AnemoiDatasetVariableWeights",
]
